"""Dask integration tests - ported from pint's test_dask.py.

Tests ArrayQuantity interop with dask arrays.
"""

from __future__ import annotations

import pytest
from pintrs import UnitRegistry
from pintrs.numpy_support import ArrayQuantity

np = pytest.importorskip("numpy", reason="numpy is not available")
dask = pytest.importorskip("dask", reason="dask is not available")
da = pytest.importorskip("dask.array", reason="dask.array is not available")


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry(force_ndarray_like=True)


class TestDaskArrayQuantityInterop:
    """Test that ArrayQuantity works with dask-computed numpy arrays."""

    def test_dask_computed_to_array_quantity(self, ureg: UnitRegistry) -> None:
        darr = da.arange(0, 10, chunks=5, dtype=float)
        nparr = darr.compute()
        q = ArrayQuantity(nparr, "meter", ureg)
        assert q.shape == (10,)
        assert q.magnitude[0] == pytest.approx(0.0)
        assert q.magnitude[9] == pytest.approx(9.0)

    def test_operations_on_computed_dask(self, ureg: UnitRegistry) -> None:
        darr = da.arange(0, 5, chunks=5, dtype=float)
        nparr = darr.compute()
        q = ArrayQuantity(nparr, "meter", ureg)
        result = q.to("kilometer")
        assert result.magnitude[1] == pytest.approx(0.001)

    def test_dask_compute_then_sum(self, ureg: UnitRegistry) -> None:
        darr = da.ones(10, chunks=5, dtype=float)
        nparr = darr.compute()
        q = ArrayQuantity(nparr, "kilogram", ureg)
        total = np.sum(q)
        assert total.magnitude == pytest.approx(10.0)

    def test_dask_compute_then_multiply(self, ureg: UnitRegistry) -> None:
        darr = da.arange(1, 6, chunks=5, dtype=float)
        nparr = darr.compute()
        meters = ArrayQuantity(nparr, "meter", ureg)
        seconds = ArrayQuantity(nparr, "second", ureg)
        result = meters * seconds
        assert result.units_str == "m * s"
        assert result.magnitude[0] == pytest.approx(1.0)

    def test_dask_2d_array(self, ureg: UnitRegistry) -> None:
        darr = da.arange(0, 25, chunks=5, dtype=float).reshape((5, 5))
        nparr = darr.compute()
        q = ArrayQuantity(nparr, "meter", ureg)
        assert q.shape == (5, 5)
        assert q.ndim == 2

    def test_force_ndarray_like_accepted(self) -> None:
        ureg = UnitRegistry(force_ndarray_like=True)
        assert ureg.force_ndarray_like is True

    def test_dask_compute_then_dot(self, ureg: UnitRegistry) -> None:
        a = da.arange(1, 4, chunks=3, dtype=float).compute()
        b = da.arange(4, 7, chunks=3, dtype=float).compute()
        qa = ArrayQuantity(a, "meter", ureg)
        qb = ArrayQuantity(b, "second", ureg)
        result = np.dot(qa, qb)
        expected = np.dot(a, b)
        assert result.magnitude == pytest.approx(expected)


class TestDaskLazyArrays:
    """Test dask lazy array operations that don't require pint wrapping."""

    def test_dask_add_then_convert(self, ureg: UnitRegistry) -> None:
        a = da.ones(5, chunks=5, dtype=float)
        b = da.full(5, 2.0, chunks=5)
        result = (a + b).compute()
        q = ArrayQuantity(result, "meter", ureg)
        assert np.all(q.magnitude == pytest.approx(3.0))

    def test_dask_reduction_then_quantity(self, ureg: UnitRegistry) -> None:
        darr = da.arange(10, chunks=5, dtype=float)
        mean_val = darr.mean().compute()
        q = ureg.Quantity(float(mean_val), "kelvin")
        assert q.magnitude == pytest.approx(4.5)



"""Pandas ExtensionArray integration for pintrs.

Provides PintDtype and PintArray so that pandas Series and DataFrames
can hold unit-aware columns, similar to pint-pandas.
"""

# pyright: basic, reportAssignmentType=false, reportAttributeAccessIssue=false, reportInvalidTypeForm=false, reportOptionalMemberAccess=false, reportReturnType=false

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    import builtins
    from collections.abc import Sequence

    import numpy as np
    from numpy.typing import DTypeLike, NDArray

    from pintrs._core import UnitRegistry as _UnitRegistry

try:
    import numpy as np
    import pandas as pd  # type: ignore[import-untyped]  # noqa: F401
    from pandas.api.extensions import (  # type: ignore[import-untyped]
        ExtensionArray,
        ExtensionDtype,
        register_extension_dtype,
    )

    has_pandas = True
except ImportError:
    has_pandas = False

if has_pandas:
    from pintrs._core import UnitRegistry as _UnitRegistry

    class PintDtype(ExtensionDtype):  # type: ignore[misc]
        """A pandas ExtensionDtype for unit-aware data.

        Args:
            units: Unit string (e.g. "meter", "kg/s").
            registry: Optional UnitRegistry. A default is created if omitted.
        """

        type = float
        na_value = float("nan")

        def __init__(
            self,
            units: str = "dimensionless",
            registry: _UnitRegistry | None = None,
        ) -> None:
            self._units = units
            self._registry = registry or _UnitRegistry()

        @property
        def name(self) -> str:
            return f"pint[{self._units}]"

        @property
        def units(self) -> str:
            return self._units

        @property
        def registry(self) -> _UnitRegistry:
            return self._registry

        @classmethod
        def construct_array_type(cls) -> builtins.type[PintArray]:
            """Return the array type associated with this dtype."""
            return PintArray

        @classmethod
        def construct_from_string(cls, string: str) -> PintDtype:
            """Create a PintDtype from a string like 'pint[meter]'."""
            if not isinstance(string, str):
                msg = f"Expected string, got {type(string)}"
                raise TypeError(msg)
            if string.startswith("pint[") and string.endswith("]"):
                units = string[5:-1]
                return PintDtype(units)
            msg = f"Cannot construct PintDtype from '{string}'"
            raise TypeError(msg)

        def __repr__(self) -> str:
            return f"PintDtype('{self._units}')"

        def __str__(self) -> str:
            return self.name

        def __hash__(self) -> int:
            return hash(self.name)

        def __eq__(self, other: object) -> bool:
            if isinstance(other, PintDtype):
                return self._units == other._units
            if isinstance(other, str):
                return self.name == other
            return NotImplemented

    class PintArray(ExtensionArray):  # type: ignore[misc]
        """A pandas ExtensionArray that holds data with physical units.

        Args:
            values: Numeric array data.
            dtype: PintDtype specifying units.
            copy: Whether to copy the data.
        """

        def __init__(
            self,
            values: NDArray[Any] | Sequence[float],
            dtype: PintDtype | None = None,
            copy: bool = False,
        ) -> None:
            if dtype is None:
                dtype = PintDtype()
            self._dtype = dtype
            arr = np.asarray(values, dtype=float)
            self._data: NDArray[Any] = arr.copy() if copy else arr

        @classmethod
        def _from_sequence(
            cls,
            scalars: Sequence[Any],
            *,
            dtype: PintDtype | None = None,
            copy: bool = False,
        ) -> PintArray:
            values = []
            inferred_units: str | None = None
            for s in scalars:
                if hasattr(s, "magnitude"):
                    values.append(s.magnitude)
                    if inferred_units is None and hasattr(s, "units"):
                        inferred_units = str(s.units)
                elif s is None or (isinstance(s, float) and np.isnan(s)):
                    values.append(float("nan"))
                else:
                    values.append(float(s))
            if dtype is None and inferred_units is not None:
                dtype = PintDtype(inferred_units)
            return PintArray(np.array(values, dtype=float), dtype=dtype, copy=copy)

        @classmethod
        def _from_factorized(
            cls,
            values: NDArray[Any],
            original: PintArray,
        ) -> PintArray:
            return PintArray(values, dtype=original._dtype)

        @property
        def dtype(self) -> PintDtype:
            return self._dtype

        @property
        def units(self) -> str:
            return self._dtype.units

        @property
        def registry(self) -> _UnitRegistry:
            return self._dtype.registry

        def __len__(self) -> int:
            return len(self._data)

        def __getitem__(self, item: Any) -> Any:
            result = self._data[item]
            if isinstance(result, np.ndarray):
                return PintArray(result, dtype=self._dtype)
            if np.isnan(result):
                return self.dtype.na_value
            return self._dtype.registry.Quantity(float(result), self._dtype.units)

        def __setitem__(self, key: Any, value: Any) -> None:
            if hasattr(value, "magnitude"):
                self._data[key] = value.magnitude
            elif value is None:
                self._data[key] = float("nan")
            else:
                self._data[key] = value

        def __repr__(self) -> str:
            return f"PintArray({self._data!r}, units='{self._dtype.units}')"

        def isna(self) -> NDArray[np.bool_]:
            result: NDArray[np.bool_] = np.isnan(self._data)
            return result

        def take(
            self,
            indices: Sequence[int],
            *,
            allow_fill: bool = False,
            fill_value: Any = None,
        ) -> PintArray:
            if allow_fill:
                fill = float("nan") if fill_value is None else float(fill_value)
                data = np.array(
                    [self._data[i] if i >= 0 else fill for i in indices],
                    dtype=float,
                )
            else:
                data = self._data[np.asarray(indices)]
            return PintArray(data, dtype=self._dtype)

        def copy(self) -> PintArray:
            return PintArray(self._data.copy(), dtype=self._dtype)

        @classmethod
        def _concat_same_type(cls, to_concat: Sequence[PintArray]) -> PintArray:
            data = np.concatenate([a._data for a in to_concat])
            return PintArray(data, dtype=to_concat[0]._dtype)

        def _values_for_factorize(self) -> tuple[NDArray[Any], float]:
            return self._data, float("nan")

        def _reduce(self, name: str, *, skipna: bool = True, **kwargs: Any) -> Any:
            funcs: dict[str, Any] = {
                "sum": np.nansum if skipna else np.sum,
                "mean": np.nanmean if skipna else np.mean,
                "median": np.nanmedian if skipna else np.median,
                "min": np.nanmin if skipna else np.min,
                "max": np.nanmax if skipna else np.max,
                "std": np.nanstd if skipna else np.std,
                "var": np.nanvar if skipna else np.var,
            }
            func = funcs.get(name)
            if func is None:
                msg = f"Reduction '{name}' is not supported"
                raise TypeError(msg)
            # numpy reduction functions don't accept pandas-specific kwargs
            kwargs.pop("min_count", None)
            result = func(self._data, **kwargs)
            return self._dtype.registry.Quantity(float(result), self._dtype.units)

        def __array__(self, dtype: DTypeLike | None = None) -> NDArray[Any]:
            if dtype is not None:
                return np.asarray(self._data, dtype=dtype)
            return self._data

        def to(self, units: str) -> PintArray:
            """Convert the array to different units.

            Args:
                units: Target unit string.
            """
            reg = self._dtype.registry
            factor = reg._get_conversion_factor(self._dtype.units, units)
            return PintArray(
                self._data * factor,
                dtype=PintDtype(units, reg),
            )

    register_extension_dtype(PintDtype)

else:

    class _PintDtypeUnavailable:
        """Placeholder that fails on use when pandas is not installed."""

        def __init__(self, *_args: object, **_kwargs: object) -> None:
            msg = (
                "Pandas support requires pandas and numpy. "
                "Install pintrs[pandas] or install pandas and numpy."
            )
            raise ModuleNotFoundError(msg)

    class _PintArrayUnavailable:
        """Placeholder that fails on use when pandas is not installed."""

        def __init__(self, *_args: object, **_kwargs: object) -> None:
            msg = (
                "Pandas support requires pandas and numpy. "
                "Install pintrs[pandas] or install pandas and numpy."
            )
            raise ModuleNotFoundError(msg)

    PintDtype: Any = _PintDtypeUnavailable  # type: ignore[no-redef]
    PintArray: Any = _PintArrayUnavailable  # type: ignore[no-redef]



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path

ROOT = Path(__file__).resolve().parents[1]


def test_import_pintrs_without_numpy_or_pandas() -> None:
    script = """
import builtins
import sys

orig_import = builtins.__import__

def blocked_import(name, globals=None, locals=None, fromlist=(), level=0):
    if name == "numpy" or name.startswith("numpy."):
        raise ImportError("blocked numpy")
    if name == "pandas" or name.startswith("pandas."):
        raise ImportError("blocked pandas")
    return orig_import(name, globals, locals, fromlist, level)

builtins.__import__ = blocked_import

import pintrs
assert pintrs.UnitRegistry is not None

try:
    pintrs.ArrayQuantity([1, 2, 3], "meter")
except ModuleNotFoundError as exc:
    assert "numpy" in str(exc).lower()
else:
    raise AssertionError("ArrayQuantity should require numpy")

try:
    pintrs.PintDtype("meter")
except ModuleNotFoundError as exc:
    assert "pandas" in str(exc).lower()
else:
    raise AssertionError("PintDtype should require pandas")
"""
    result = subprocess.run(  # noqa: S603
        [sys.executable, "-c", script],
        cwd=ROOT,
        env={**os.environ, "PYTHONPATH": str(ROOT / "python")},
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0, result.stderr



"""Babel formatting tests - ported from pint's test_babel.py."""

from __future__ import annotations

import pytest

babel = pytest.importorskip("babel")

from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestBabelFormatting:
    def test_format_default_locale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000.5, "meter")
        result = q.format_babel("en")
        assert "1" in result
        assert "000" in result or "," in result

    def test_format_german_locale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000.5, "meter")
        result = q.format_babel("de_DE")
        assert "1" in result

    def test_format_french_locale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1000.5, "meter")
        result = q.format_babel("fr_FR")
        assert "1" in result

    def test_unit_format_babel(self, ureg: UnitRegistry) -> None:
        u = ureg.parse_units("meter")
        result = u.format_babel("en")
        assert "meter" in result or "m" in result

    def test_integer_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(42, "second")
        result = q.format_babel("en")
        assert "42" in result

    def test_zero_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "meter")
        result = q.format_babel("en")
        assert "0" in result

    def test_negative_quantity(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(-5, "meter")
        result = q.format_babel("en")
        assert "-5" in result

    def test_large_number(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1_000_000, "meter")
        result = q.format_babel("en_US")
        assert "1" in result
        assert "000" in result


class TestBabelNoLocale:
    def test_no_babel_fallback(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(10, "meter")
        result = q.format_babel()
        assert "10" in result



"""Logarithmic unit support for pintrs.

Provides LogarithmicQuantity for units like dB, dBm, dBW, Np (neper), and Bel.
These units represent ratios on a logarithmic scale and require special
arithmetic rules.
"""

from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    from pintrs._core import UnitRegistry

_EQ_TOLERANCE = 1e-10


class LogarithmicQuantity:
    """A quantity expressed in logarithmic units (dB, dBm, Np, etc.).

    Logarithmic quantities follow different arithmetic rules than linear
    quantities:
    - Adding two dB values = multiplying the underlying linear values
    - A dBm value has a reference of 1 mW

    Args:
        magnitude: The value in logarithmic units.
        units: The logarithmic unit string ("dB", "dBm", "dBW", "Np", "Bel").
        registry: Optional UnitRegistry for conversions.
    """

    _REFERENCE_POWER: ClassVar[dict[str, tuple[float, str]]] = {
        "dBm": (1e-3, "watt"),
        "dBW": (1.0, "watt"),
        "dBmW": (1e-3, "watt"),
    }

    _LOG_FACTOR: ClassVar[dict[str, float]] = {
        "dB": 10.0,
        "dBm": 10.0,
        "dBW": 10.0,
        "dBmW": 10.0,
        "B": 1.0,
        "Bel": 1.0,
        "Np": 1.0,
    }

    @staticmethod
    def _canonical_units(units: str) -> str:
        return "Bel" if units == "B" else units

    def __init__(
        self,
        magnitude: float,
        units: str = "dB",
        registry: UnitRegistry | None = None,
    ) -> None:
        units = self._canonical_units(units)
        if units not in self._LOG_FACTOR:
            valid = ", ".join(self._LOG_FACTOR)
            msg = f"Unknown logarithmic unit: {units!r}. Valid: {valid}"
            raise ValueError(msg)
        self._magnitude = float(magnitude)
        self._units = units
        self._registry = registry

    @property
    def magnitude(self) -> float:
        """The logarithmic magnitude."""
        return self._magnitude

    @property
    def m(self) -> float:
        """Alias for magnitude."""
        return self._magnitude

    @property
    def units(self) -> str:
        """The logarithmic unit string."""
        return self._units

    def to_linear(self) -> float:
        """Convert to linear scale (ratio).

        For dBm/dBW this returns power in the reference unit (mW or W).
        For dB/Np this returns the dimensionless ratio.
        """
        if self._units == "Np":
            return float(math.exp(2.0 * self._magnitude))
        log_factor = self._LOG_FACTOR[self._units]
        return float(10.0 ** (self._magnitude / log_factor))

    @classmethod
    def from_linear(
        cls,
        value: float,
        units: str = "dB",
        registry: UnitRegistry | None = None,
    ) -> LogarithmicQuantity:
        """Create from a linear value.

        Args:
            value: Linear value (ratio or power).
            units: Target logarithmic unit.
            registry: Optional UnitRegistry.
        """
        units = cls._canonical_units(units)
        if value <= 0:
            msg = "Cannot convert non-positive value to logarithmic scale"
            raise ValueError(msg)
        if units == "Np":
            magnitude = 0.5 * math.log(value)
        else:
            log_factor = cls._LOG_FACTOR[units]
            magnitude = log_factor * math.log10(value)
        return cls(magnitude, units, registry)

    def to(self, units: str) -> LogarithmicQuantity:
        """Convert between logarithmic units.

        Args:
            units: Target logarithmic unit string.
        """
        units = self._canonical_units(units)
        if units == self._units:
            return LogarithmicQuantity(self._magnitude, units, self._registry)

        if units not in self._LOG_FACTOR:
            msg = f"Unknown logarithmic unit: {units!r}"
            raise ValueError(msg)

        src_ref = self._REFERENCE_POWER.get(self._units)
        dst_ref = self._REFERENCE_POWER.get(units)

        if src_ref is not None and dst_ref is not None:
            # Both have absolute references: convert through watts
            src_ref_val = src_ref[0]
            dst_ref_val = dst_ref[0]
            dst_log_factor = self._LOG_FACTOR[units]
            offset = dst_log_factor * math.log10(src_ref_val / dst_ref_val)
            return LogarithmicQuantity(
                self._magnitude + offset,
                units,
                self._registry,
            )

        # Relative or mixed: convert via linear ratio
        linear = self.to_linear()
        return LogarithmicQuantity.from_linear(linear, units, self._registry)

    def to_quantity(self) -> Any:
        """Convert to a linear Quantity (requires a reference unit like dBm/dBW).

        Returns:
            A pintrs Quantity with the linear value and reference unit.

        Raises:
            ValueError: If the unit has no absolute reference (e.g. plain dB).
        """
        ref = self._REFERENCE_POWER.get(self._units)
        if ref is None:
            msg = f"Cannot convert relative unit {self._units!r} to absolute Quantity"
            raise ValueError(msg)
        ref_value, ref_unit = ref
        linear = self.to_linear() * ref_value
        if self._registry is not None:
            return self._registry.Quantity(linear, ref_unit)
        from pintrs._core import Quantity  # noqa: PLC0415

        return Quantity(linear, ref_unit)

    def __repr__(self) -> str:
        return f"<LogarithmicQuantity({self._magnitude}, '{self._units}')>"

    def __str__(self) -> str:
        return f"{self._magnitude} {self._units}"

    def __add__(self, other: LogarithmicQuantity | float) -> LogarithmicQuantity:
        if isinstance(other, LogarithmicQuantity):
            if other._units != self._units:
                other = other.to(self._units)
            return LogarithmicQuantity(
                self._magnitude + other._magnitude,
                self._units,
                self._registry,
            )
        return LogarithmicQuantity(
            self._magnitude + float(other),
            self._units,
            self._registry,
        )

    def __radd__(self, other: float) -> LogarithmicQuantity:
        return self.__add__(other)

    def __sub__(self, other: LogarithmicQuantity | float) -> LogarithmicQuantity:
        if isinstance(other, LogarithmicQuantity):
            if other._units != self._units:
                other = other.to(self._units)
            return LogarithmicQuantity(
                self._magnitude - other._magnitude,
                self._units,
                self._registry,
            )
        return LogarithmicQuantity(
            self._magnitude - float(other),
            self._units,
            self._registry,
        )

    def __rsub__(self, other: float) -> LogarithmicQuantity:
        return LogarithmicQuantity(
            float(other) - self._magnitude,
            self._units,
            self._registry,
        )

    def __neg__(self) -> LogarithmicQuantity:
        return LogarithmicQuantity(-self._magnitude, self._units, self._registry)

    def __mul__(self, other: float) -> LogarithmicQuantity:
        return LogarithmicQuantity(
            self._magnitude * float(other),
            self._units,
            self._registry,
        )

    def __rmul__(self, other: float) -> LogarithmicQuantity:
        return self.__mul__(other)

    def __truediv__(self, other: float) -> LogarithmicQuantity:
        return LogarithmicQuantity(
            self._magnitude / float(other),
            self._units,
            self._registry,
        )

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LogarithmicQuantity):
            if other._units != self._units:
                other = other.to(self._units)
            return abs(self._magnitude - other._magnitude) < _EQ_TOLERANCE
        return NotImplemented

    def __lt__(self, other: LogarithmicQuantity) -> bool:
        if other._units != self._units:
            other = other.to(self._units)
        return self._magnitude < other._magnitude

    def __le__(self, other: LogarithmicQuantity) -> bool:
        if other._units != self._units:
            other = other.to(self._units)
        return self._magnitude <= other._magnitude

    def __gt__(self, other: LogarithmicQuantity) -> bool:
        if other._units != self._units:
            other = other.to(self._units)
        return self._magnitude > other._magnitude

    def __ge__(self, other: LogarithmicQuantity) -> bool:
        if other._units != self._units:
            other = other.to(self._units)
        return self._magnitude >= other._magnitude

    def __hash__(self) -> int:
        return hash((round(self._magnitude, 10), self._units))



"""Basic pintrs usage: quantities, conversions, and arithmetic."""

from __future__ import annotations

from pintrs import Measurement, Quantity, UnitRegistry

ureg = UnitRegistry()

# --- Creating quantities ---

distance = ureg.Quantity(100.0, "kilometer")
time = ureg.Quantity(1.5, "hour")
print(f"Distance: {distance}")
print(f"Time: {time}")

# From string expression
q = Quantity("9.81 m/s**2")
print(f"Gravity: {q}")

# Attribute-style
print(f"ureg.meter: {ureg.meter}")
print(f"ureg.speed_of_light: {ureg.speed_of_light}")

# --- Conversions ---

print(f"\n{distance} = {distance.to('mile')}")
print(f"{distance} = {distance.to('meter')}")
print(f"{time} = {time.to('minute')}")

# In-place conversion
speed = distance / time
print(f"\nSpeed: {speed}")
print(f"Speed in m/s: {speed.to('m/s')}")
print(f"Speed compact: {speed.to_compact()}")

# --- Arithmetic ---

d1 = ureg.Quantity(10.0, "meter")
d2 = ureg.Quantity(5.0, "meter")
print(f"\n{d1} + {d2} = {d1 + d2}")
print(f"{d1} - {d2} = {d1 - d2}")
print(f"{d1} * 3 = {d1 * 3}")
print(f"{d1} / 2 = {d1 / 2}")
print(f"{d1} ** 2 = {d1**2}")

# Mixed units
d3 = ureg.Quantity(1.0, "kilometer")
print(f"\n{d1} + {d3} = {d1 + d3}")

# Unit multiplication
area = d1 * ureg.Quantity(4.0, "meter")
print(f"Area: {area}")
print(f"Area in base units: {area.to_base_units()}")

# --- Comparisons ---

print(f"\n{d1} > {d2}: {d1 > d2}")
print(f"{d1} == {d2}: {d1 == d2}")

# --- Properties ---

print(f"\nDimensionality of {speed}: {speed.dimensionality}")
print(f"Is {speed} dimensionless? {speed.dimensionless}")
print(f"Compatible units for meter: {ureg.Quantity(1, 'meter').compatible_units()[:5]}")

# --- Measurements (uncertainty) ---

m = Measurement(Quantity(100.0, "meter"), 0.5)
print(f"\nMeasurement: {m}")
print(f"Relative error: {m.rel:.4f}")

m2 = Measurement(Quantity(50.0, "meter"), 0.3)
combined = m + m2
print(f"{m} + {m2} = {combined}")

# --- Custom units ---

ureg.define("smoot = 1.7018 * meter")
bridge = ureg.Quantity(364.4, "smoot")
print(f"\nHarvard Bridge: {bridge}")
print(f"In meters: {bridge.to('meter')}")



"""Tests for pintrs System functionality."""

from __future__ import annotations

import pintrs
from pintrs.system import System


def test_system_creation():
    sys = System("test_sys_1")
    assert sys.name == "test_sys_1"
    assert sys.base_units == {}


def test_system_add_remove_rule():
    sys = System("test_sys_2")
    sys.add_rule("[length]", "meter")
    sys.add_rule("[mass]", "kilogram")
    assert sys.base_units == {"[length]": "meter", "[mass]": "kilogram"}

    sys.remove_rule("[mass]")
    assert sys.base_units == {"[length]": "meter"}


def test_system_preferred_unit_for():
    sys = System("test_sys_3")
    sys.add_rule("[length]", "meter")
    assert sys.preferred_unit_for("[length]") == "meter"
    assert sys.preferred_unit_for("[mass]") is None


def test_system_contains():
    sys = System("test_sys_4")
    sys.add_rule("[length]", "meter")
    assert "[length]" in sys
    assert "[mass]" not in sys


def test_system_registry_lookup():
    sys = System("test_sys_lookup")
    assert System.get("test_sys_lookup") is sys
    assert System.get("nonexistent_sys") is None


def test_system_from_lines():
    lines = ["test_from_lines_sys", "[length]: meter", "[mass]: kilogram"]
    sys = System.from_lines(lines)
    assert sys.name == "test_from_lines_sys"
    assert sys.base_units == {"[length]": "meter", "[mass]": "kilogram"}


def test_system_repr():
    sys = System("test_sys_repr")
    sys.add_rule("[length]", "meter")
    assert "test_sys_repr" in repr(sys)
    assert "1" in repr(sys)


def test_system_convert():
    ureg = pintrs.UnitRegistry()
    cgs = ureg.get_system("cgs")
    val, unit = cgs.convert(ureg, 1.0, "meter")
    assert abs(val - 100.0) < 1e-10
    assert unit == "centimeter"


def test_system_convert_mass():
    ureg = pintrs.UnitRegistry()
    cgs = ureg.get_system("cgs")
    val, unit = cgs.convert(ureg, 1.0, "kilogram")
    assert abs(val - 1000.0) < 1e-10
    assert unit == "gram"


def test_system_convert_no_rule():
    ureg = pintrs.UnitRegistry()
    sys = System("test_sys_norule")
    val, unit = sys.convert(ureg, 42.0, "meter")
    assert val == 42.0
    assert unit == "meter"


def test_builtin_mks():
    ureg = pintrs.UnitRegistry()
    mks = ureg.get_system("mks")
    assert mks.preferred_unit_for("[length]") == "meter"
    assert mks.preferred_unit_for("[mass]") == "kilogram"
    assert mks.preferred_unit_for("[time]") == "second"


def test_builtin_si():
    ureg = pintrs.UnitRegistry()
    si = ureg.get_system("SI")
    assert si.preferred_unit_for("[length]") == "meter"
    assert si.preferred_unit_for("[mass]") == "kilogram"


def test_builtin_cgs():
    ureg = pintrs.UnitRegistry()
    cgs = ureg.get_system("cgs")
    assert cgs.preferred_unit_for("[length]") == "centimeter"
    assert cgs.preferred_unit_for("[mass]") == "gram"


def test_builtin_imperial():
    ureg = pintrs.UnitRegistry()
    imp = ureg.get_system("imperial")
    assert imp.preferred_unit_for("[length]") == "foot"
    assert imp.preferred_unit_for("[mass]") == "pound"


def test_builtin_atomic():
    ureg = pintrs.UnitRegistry()
    atomic = ureg.get_system("atomic")
    assert atomic.preferred_unit_for("[length]") == "bohr"
    assert atomic.preferred_unit_for("[mass]") == "electron_mass"


def test_get_unknown_system_creates_empty():
    ureg = pintrs.UnitRegistry()
    sys = ureg.get_system("brand_new_system")
    assert sys.name == "brand_new_system"
    assert sys.base_units == {}


def test_mks_convert_foot_to_meter():
    ureg = pintrs.UnitRegistry()
    mks = ureg.get_system("mks")
    val, unit = mks.convert(ureg, 1.0, "foot")
    assert abs(val - 0.3048) < 1e-10
    assert unit == "meter"



"""NumPy integration tests - ported from pint's test_numpy.py."""

from __future__ import annotations

import math

import pytest

np = pytest.importorskip("numpy")

from pintrs import DimensionalityError, UnitRegistry
from pintrs.numpy_support import ArrayQuantity


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestArrayQuantityCreation:
    def test_from_list(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        assert len(aq.magnitude) == 3
        np.testing.assert_array_equal(aq.magnitude, [1, 2, 3])

    def test_from_ndarray(self, ureg: UnitRegistry) -> None:
        arr = np.array([1.0, 2.0, 3.0])
        aq = ArrayQuantity(arr, "meter", ureg)
        np.testing.assert_array_equal(aq.magnitude, arr)

    def test_units_str(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert "meter" in str(aq.units_str) or "m" in str(aq.units_str)

    def test_shape(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.zeros((2, 3)), "meter", ureg)
        assert aq.magnitude.shape == (2, 3)

    def test_dtype(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0], "meter", ureg)
        assert aq.magnitude.dtype == np.float64

    def test_empty_array(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([], "meter", ureg)
        assert len(aq.magnitude) == 0


class TestArrayQuantityConversion:
    def test_to(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1000, 2000], "meter", ureg)
        result = aq.to("kilometer")
        np.testing.assert_allclose(result.magnitude, [1, 2])

    def test_to_base_units(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "kilometer", ureg)
        result = aq.to_base_units()
        np.testing.assert_allclose(result.magnitude, [1000, 2000])

    def test_incompatible_raises(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        with pytest.raises(DimensionalityError):
            aq.to("second")


class TestArrayQuantityArithmetic:
    def test_add_same_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([3, 4], "meter", ureg)
        result = a + b
        np.testing.assert_array_equal(result.magnitude, [4, 6])

    def test_add_compatible_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([100, 200], "centimeter", ureg)
        result = a + b
        np.testing.assert_allclose(result.magnitude, [2, 4])

    def test_sub(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([5, 6], "meter", ureg)
        b = ArrayQuantity([1, 2], "meter", ureg)
        result = a - b
        np.testing.assert_array_equal(result.magnitude, [4, 4])

    def test_mul_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        result = a * 3
        np.testing.assert_array_equal(result.magnitude, [3, 6])

    def test_mul_array_quantities(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        b = ArrayQuantity([4, 5], "second", ureg)
        result = a * b
        np.testing.assert_array_equal(result.magnitude, [8, 15])

    def test_div_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([6, 8], "meter", ureg)
        result = a / 2
        np.testing.assert_array_equal(result.magnitude, [3, 4])

    def test_div_array_quantities(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([10, 20], "meter", ureg)
        b = ArrayQuantity([2, 4], "second", ureg)
        result = a / b
        np.testing.assert_array_equal(result.magnitude, [5, 5])

    def test_pow(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        result = a**2
        np.testing.assert_array_equal(result.magnitude, [4, 9])

    def test_neg(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, -2], "meter", ureg)
        result = -a
        np.testing.assert_array_equal(result.magnitude, [-1, 2])

    def test_abs(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([-1, 2, -3], "meter", ureg)
        result = abs(a)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 3])


class TestArrayQuantityComparison:
    def test_eq(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([1, 2], "meter", ureg)
        assert a == b

    def test_ne(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([1, 3], "meter", ureg)
        assert a != b


class TestArrayQuantityRepresentation:
    def test_str(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        s = str(aq)
        assert "1" in s
        assert "meter" in s or "m" in s

    def test_repr(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        r = repr(aq)
        assert "ArrayQuantity" in r or "1" in r


class TestArrayUfunc:
    def test_add_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([3, 4], "meter", ureg)
        result = np.add(a, b)
        np.testing.assert_array_equal(result.magnitude, [4, 6])

    def test_multiply_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        result = np.multiply(a, 2)
        np.testing.assert_array_equal(result.magnitude, [4, 6])

    def test_sqrt_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([4, 9], "meter**2", ureg)
        result = np.sqrt(a)
        np.testing.assert_allclose(result.magnitude, [2, 3])

    def test_sin_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, math.pi / 2], "radian", ureg)
        result = np.sin(a)
        np.testing.assert_allclose(result.magnitude, [0, 1], atol=1e-10)

    def test_sum(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.sum(a)
        assert float(result.magnitude) == pytest.approx(6.0)

    def test_mean(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 4, 6], "meter", ureg)
        result = np.mean(a)
        assert float(result.magnitude) == pytest.approx(4.0)


class TestArrayQuantityLikeCreation:
    def test_ones_like(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.ones_like(a.magnitude)
        np.testing.assert_array_equal(result, [1, 1, 1])

    def test_zeros_like(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.zeros_like(a.magnitude)
        np.testing.assert_array_equal(result, [0, 0, 0])


class TestArrayQuantityShapeOps:
    def test_flatten(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        flat = a.magnitude.flatten()
        np.testing.assert_array_equal(flat, [1, 2, 3, 4])

    def test_reshape(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        reshaped = a.magnitude.reshape(2, 2)
        assert reshaped.shape == (2, 2)

    def test_transpose(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        t = a.magnitude.T
        np.testing.assert_array_equal(t, [[1, 3], [2, 4]])


class TestArrayQuantityReduction:
    def test_min(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([3, 1, 2], "meter", ureg)
        assert np.min(a.magnitude) == 1

    def test_max(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([3, 1, 2], "meter", ureg)
        assert np.max(a.magnitude) == 3

    def test_argmin(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([3, 1, 2], "meter", ureg)
        assert np.argmin(a.magnitude) == 1

    def test_argmax(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([3, 1, 2], "meter", ureg)
        assert np.argmax(a.magnitude) == 0

    def test_cumsum(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.cumsum(a.magnitude)
        np.testing.assert_array_equal(result, [1, 3, 6])

    def test_prod(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3, 4], "dimensionless", ureg)
        assert np.prod(a.magnitude) == 24

    def test_std(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 4, 4, 4, 5, 5, 7, 9], "meter", ureg)
        assert np.std(a.magnitude) == pytest.approx(2.0)

    def test_var(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 4, 4, 4, 5, 5, 7, 9], "meter", ureg)
        assert np.var(a.magnitude) == pytest.approx(4.0)


class TestQuantityFromList:
    def test_from_list(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(2, "meter")
        from pintrs import Quantity

        result = Quantity.from_list([q1, q2])
        assert result.magnitude[0] == pytest.approx(1.0)
        assert result.magnitude[1] == pytest.approx(2.0)

    def test_from_sequence(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(2, "meter")
        from pintrs import Quantity

        result = Quantity.from_sequence(iter([q1, q2]))
        assert result.magnitude[0] == pytest.approx(1.0)
        assert result.magnitude[1] == pytest.approx(2.0)

    def test_from_list_with_units(self, ureg: UnitRegistry) -> None:
        from pintrs import Quantity

        result = Quantity.from_list(
            [ureg.Quantity(100, "cm"), ureg.Quantity(2, "meter")], units="meter"
        )
        np.testing.assert_allclose(result.magnitude, [1.0, 2.0])


class TestArrayQuantityIndexingSlicing:
    """Tests for __getitem__, __setitem__, slicing, and fancy indexing."""

    def test_getitem_scalar(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([10, 20, 30], "meter", ureg)
        result = aq[1]
        assert float(result.magnitude) == pytest.approx(20.0)

    def test_getitem_slice(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([10, 20, 30, 40], "meter", ureg)
        result = aq[1:3]
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [20, 30])

    def test_getitem_2d_row(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        row = aq[0]
        assert isinstance(row, ArrayQuantity)
        np.testing.assert_array_equal(row.magnitude, [1, 2])

    def test_getitem_2d_element(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        elem = aq.magnitude[1, 1]
        assert elem == 4

    def test_getitem_boolean_mask(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        mask = np.array([True, False, True, False])
        result = aq[mask]
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [1, 3])

    def test_getitem_fancy_index(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([10, 20, 30, 40], "meter", ureg)
        result = aq[np.array([0, 3])]
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [10, 40])

    def test_setitem_scalar(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        aq[0] = 99.0
        assert aq.magnitude[0] == pytest.approx(99.0)

    def test_setitem_slice(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        aq[1:3] = 0.0
        np.testing.assert_array_equal(aq.magnitude, [1.0, 0.0, 0.0, 4.0])

    def test_setitem_with_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        aq[0] = ureg.Quantity(99.0, "meter")
        assert aq.magnitude[0] == pytest.approx(99.0)

    def test_setitem_with_array_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        aq[1:3] = ArrayQuantity([10.0, 20.0], "meter", ureg)
        np.testing.assert_array_equal(aq.magnitude, [1.0, 10.0, 20.0, 4.0])

    def test_setitem_ellipsis(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        aq[...] = 7.0
        np.testing.assert_array_equal(aq.magnitude, [7.0, 7.0, 7.0])


class TestArrayQuantityIteration:
    """Tests for __iter__ and __len__."""

    def test_iter_1d(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        values = list(aq)
        assert len(values) == 4
        for i, v in enumerate(values):
            assert float(v.magnitude) == pytest.approx(float(i + 1))

    def test_iter_2d(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        rows = list(aq)
        assert len(rows) == 2
        assert isinstance(rows[0], ArrayQuantity)
        np.testing.assert_array_equal(rows[0].magnitude, [1, 2])
        np.testing.assert_array_equal(rows[1].magnitude, [3, 4])

    def test_len(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        assert len(aq) == 3

    def test_len_2d(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.zeros((3, 4)), "meter", ureg)
        assert len(aq) == 3


class TestArrayQuantityComparisonOps:
    """Tests for comparison operators (lt, le, gt, ge) with units."""

    def test_lt_same_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = a < ArrayQuantity([2, 2, 4, 4], "meter", ureg)
        np.testing.assert_array_equal(result, [True, False, True, False])

    def test_gt_same_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = a > ArrayQuantity([0, 2, 2, 5], "meter", ureg)
        np.testing.assert_array_equal(result, [True, False, True, False])

    def test_le(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = a <= ArrayQuantity([1, 3, 2], "meter", ureg)
        np.testing.assert_array_equal(result, [True, True, False])

    def test_ge(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = a >= ArrayQuantity([1, 3, 2], "meter", ureg)
        np.testing.assert_array_equal(result, [True, False, True])

    def test_lt_compatible_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([50, 300], "centimeter", ureg)
        result = a < b
        np.testing.assert_array_equal(result, [False, True])

    def test_gt_compatible_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([50, 300], "centimeter", ureg)
        result = a > b
        np.testing.assert_array_equal(result, [True, False])

    def test_comparison_with_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = a > 2
        np.testing.assert_array_equal(result, [False, False, True, True])


class TestArrayUfuncPower:
    """Tests for power, square, cbrt ufuncs."""

    def test_power_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        result = np.power(a, 2)
        np.testing.assert_array_equal(result.magnitude, [4, 9])

    def test_power_ufunc_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        result = np.power(a, 2)
        expected_units = str(ureg.Unit("meter") ** 2)
        assert str(result.units) == expected_units

    def test_square_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        result = np.square(a)
        np.testing.assert_array_equal(result.magnitude, [4, 9])

    def test_cbrt_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([8, 27], "meter**3", ureg)
        result = np.cbrt(a)
        np.testing.assert_allclose(result.magnitude, [2, 3])


class TestArrayUfuncExpLog:
    """Tests for exp, log ufuncs requiring dimensionless input."""

    def test_exp_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, 1, 2], "dimensionless", ureg)
        result = np.exp(a)
        np.testing.assert_allclose(result.magnitude, [1.0, math.e, math.e**2])

    def test_exp_with_units_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        with pytest.raises(DimensionalityError):
            np.exp(a)

    def test_log_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, math.e, math.e**2], "dimensionless", ureg)
        result = np.log(a)
        np.testing.assert_allclose(result.magnitude, [0, 1, 2])

    def test_log_with_units_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        with pytest.raises(DimensionalityError):
            np.log(a)

    def test_log10_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 10, 100], "dimensionless", ureg)
        result = np.log10(a)
        np.testing.assert_allclose(result.magnitude, [0, 1, 2])

    def test_log2_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 4], "dimensionless", ureg)
        result = np.log2(a)
        np.testing.assert_allclose(result.magnitude, [0, 1, 2])

    def test_exp2_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, 1, 3], "dimensionless", ureg)
        result = np.exp2(a)
        np.testing.assert_allclose(result.magnitude, [1, 2, 8])

    def test_exp2_with_units_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        with pytest.raises(DimensionalityError):
            np.exp2(a)


class TestArrayUfuncTrig:
    """Tests for trigonometric ufuncs with unit handling."""

    def test_cos_radians(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, math.pi], "radian", ureg)
        result = np.cos(a)
        np.testing.assert_allclose(result.magnitude, [1, -1], atol=1e-10)

    def test_tan_radians(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, math.pi / 4], "radian", ureg)
        result = np.tan(a)
        np.testing.assert_allclose(result.magnitude, [0, 1], atol=1e-10)

    def test_sin_with_units_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        with pytest.raises(DimensionalityError):
            np.sin(a)

    def test_arcsin_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, 1], "dimensionless", ureg)
        result = np.arcsin(a)
        np.testing.assert_allclose(result.magnitude, [0, math.pi / 2], atol=1e-10)

    def test_arccos_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 0], "dimensionless", ureg)
        result = np.arccos(a)
        np.testing.assert_allclose(result.magnitude, [0, math.pi / 2], atol=1e-10)

    def test_arctan_dimensionless(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0, 1], "dimensionless", ureg)
        result = np.arctan(a)
        np.testing.assert_allclose(result.magnitude, [0, math.pi / 4], atol=1e-10)

    def test_arcsin_with_units_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([0.5], "meter", ureg)
        with pytest.raises(DimensionalityError):
            np.arcsin(a)


class TestArrayUfuncComparison:
    """Tests for comparison ufuncs (np.greater, np.less, etc.)."""

    def test_greater(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([2, 2, 2], "meter", ureg)
        result = np.greater(a, b)
        np.testing.assert_array_equal(result, [False, False, True])

    def test_less(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([2, 2, 2], "meter", ureg)
        result = np.less(a, b)
        np.testing.assert_array_equal(result, [True, False, False])

    def test_equal_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([1, 5, 3], "meter", ureg)
        result = np.equal(a, b)
        np.testing.assert_array_equal(result, [True, False, True])

    def test_not_equal_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([1, 5, 3], "meter", ureg)
        result = np.not_equal(a, b)
        np.testing.assert_array_equal(result, [False, True, False])

    def test_greater_equal(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([2, 2, 2], "meter", ureg)
        result = np.greater_equal(a, b)
        np.testing.assert_array_equal(result, [False, True, True])

    def test_less_equal(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ArrayQuantity([2, 2, 2], "meter", ureg)
        result = np.less_equal(a, b)
        np.testing.assert_array_equal(result, [True, True, False])

    def test_maximum_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 4], "meter", ureg)
        b = ArrayQuantity([3, 2], "meter", ureg)
        result = np.maximum(a, b)
        np.testing.assert_array_equal(result.magnitude, [3, 4])

    def test_minimum_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 4], "meter", ureg)
        b = ArrayQuantity([3, 2], "meter", ureg)
        result = np.minimum(a, b)
        np.testing.assert_array_equal(result.magnitude, [1, 2])


class TestArrayQuantityBroadcasting:
    """Tests for broadcasting behavior with quantities."""

    def test_add_broadcast_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        b = ureg.Quantity(10, "meter")
        result = a + b
        np.testing.assert_array_equal(result.magnitude, [11, 12, 13])

    def test_mul_broadcast_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = a * 5
        np.testing.assert_array_equal(result.magnitude, [5, 10, 15])

    def test_rmul_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = 5 * a
        np.testing.assert_array_equal(result.magnitude, [5, 10, 15])

    def test_sub_broadcast_scalar_quantity(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([10, 20, 30], "meter", ureg)
        b = ureg.Quantity(5, "meter")
        result = a - b
        np.testing.assert_array_equal(result.magnitude, [5, 15, 25])

    def test_div_broadcast_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([10, 20, 30], "meter", ureg)
        result = a / 10
        np.testing.assert_allclose(result.magnitude, [1, 2, 3])

    def test_rdiv_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 4], "meter", ureg)
        result = 8 / a
        np.testing.assert_allclose(result.magnitude, [8, 4, 2])


class TestArrayQuantityNanHandling:
    """Tests for NaN handling in array quantities."""

    def test_nan_in_array(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        assert np.isnan(aq.magnitude[1])

    def test_isnan_ufunc(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        result = np.isnan(aq)
        np.testing.assert_array_equal(result, [False, True, False])

    def test_isinf_ufunc(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.inf, -np.inf], "meter", ureg)
        result = np.isinf(aq)
        np.testing.assert_array_equal(result, [False, True, True])

    def test_isfinite_ufunc(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, np.inf], "meter", ureg)
        result = np.isfinite(aq)
        np.testing.assert_array_equal(result, [True, False, False])

    def test_nan_sum(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        result = np.nansum(aq.magnitude)
        assert result == pytest.approx(4.0)

    def test_nan_mean(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        result = np.nanmean(aq.magnitude)
        assert result == pytest.approx(2.0)

    def test_nan_min(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        assert np.nanmin(aq.magnitude) == pytest.approx(1.0)

    def test_nan_max(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, np.nan, 3], "meter", ureg)
        assert np.nanmax(aq.magnitude) == pytest.approx(3.0)


class TestArrayQuantityConcatenation:
    """Tests for np.concatenate and np.stack with array quantities."""

    def test_concatenate_same_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([3, 4], "meter", ureg)
        result = np.concatenate([a, b])
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 3, 4])

    def test_concatenate_2d(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity(np.array([[1, 2]]), "meter", ureg)
        b = ArrayQuantity(np.array([[3, 4]]), "meter", ureg)
        result = np.concatenate([a, b], axis=0)
        np.testing.assert_array_equal(result.magnitude, [[1, 2], [3, 4]])

    def test_stack_same_units(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([3, 4], "meter", ureg)
        result = np.stack([a, b])
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [[1, 2], [3, 4]])


class TestArrayQuantityWhere:
    """Tests for np.where with array quantities."""

    def test_where_basic(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        b = ArrayQuantity([10, 20, 30, 40], "meter", ureg)
        cond = np.array([True, False, True, False])
        result = ArrayQuantity(np.where(cond, a.magnitude, b.magnitude), "meter", ureg)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [1, 20, 3, 40])

    def test_where_with_scalar(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        cond = np.array([True, False, True, False])
        result = ArrayQuantity(np.where(cond, a.magnitude, 0), "meter", ureg)
        np.testing.assert_array_equal(result.magnitude, [1, 0, 3, 0])

    def test_where_from_comparison(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        b = ArrayQuantity([0, 0, 0, 0], "meter", ureg)
        cond = a.magnitude >= 3
        result = ArrayQuantity(np.where(cond, a.magnitude, b.magnitude), "meter", ureg)
        np.testing.assert_array_equal(result.magnitude, [0, 0, 3, 4])

    def test_where_magnitude_roundtrip(self, ureg: UnitRegistry) -> None:
        """Test np.where via magnitude extraction and re-wrapping."""
        a = ArrayQuantity([10, 20, 30, 40], "meter", ureg)
        b = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        cond = np.array([True, True, False, False])
        result = ArrayQuantity(np.where(cond, a.magnitude, b.magnitude), "meter", ureg)
        np.testing.assert_array_equal(result.magnitude, [10, 20, 3, 4])


class TestArrayQuantityClip:
    """Tests for clip operations."""

    def test_clip_min(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4, 5], "meter", ureg)
        result = aq.clip(min=3)
        np.testing.assert_array_equal(result.magnitude, [3, 3, 3, 4, 5])

    def test_clip_max(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4, 5], "meter", ureg)
        result = aq.clip(max=3)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 3, 3, 3])

    def test_clip_both(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4, 5], "meter", ureg)
        result = aq.clip(min=2, max=4)
        np.testing.assert_array_equal(result.magnitude, [2, 2, 3, 4, 4])

    def test_np_clip(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4, 5], "meter", ureg)
        result = np.clip(aq, 2, 4)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [2, 2, 3, 4, 4])


class TestArrayQuantitySorting:
    """Tests for sort, argsort operations."""

    def test_np_sort(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([4, 1, 3, 2], "meter", ureg)
        result = np.sort(aq)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 3, 4])

    def test_np_argsort(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([4, 1, 3, 2], "meter", ureg)
        result = np.argsort(aq)
        np.testing.assert_array_equal(result, [1, 3, 2, 0])

    def test_np_sort_2d(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[3, 1], [4, 2]]), "meter", ureg)
        result = np.sort(aq, axis=1)
        np.testing.assert_array_equal(result.magnitude, [[1, 3], [2, 4]])


class TestArrayQuantityFillPutSearchsorted:
    """Tests for fill, put, searchsorted methods."""

    def test_fill_scalar(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        aq.fill(7.0)
        np.testing.assert_array_equal(aq.magnitude, [7.0, 7.0, 7.0])

    def test_fill_with_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        aq.fill(ureg.Quantity(5.0, "meter"))
        np.testing.assert_array_equal(aq.magnitude, [5.0, 5.0, 5.0])

    def test_put_scalar(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        aq.put(np.array([0, 2]), 99.0)
        np.testing.assert_array_equal(aq.magnitude, [99.0, 2.0, 99.0, 4.0])

    def test_put_with_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        aq.put(
            np.array([0, 2]),
            ArrayQuantity([10.0, 20.0], "meter", ureg),
        )
        np.testing.assert_array_equal(aq.magnitude, [10.0, 2.0, 20.0, 4.0])

    def test_searchsorted_basic(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        result = aq.searchsorted(2.5)
        assert result == 2

    def test_searchsorted_with_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        result = aq.searchsorted(ureg.Quantity(2.5, "meter"))
        assert result == 2

    def test_searchsorted_array(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0, 4.0], "meter", ureg)
        result = aq.searchsorted(
            ArrayQuantity([1.5, 3.5], "meter", ureg),
        )
        np.testing.assert_array_equal(result, [1, 3])


class TestArrayQuantityComplex:
    """Tests for complex number handling."""

    def test_complex_creation(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 2j, 3 + 4j], "meter", ureg)
        assert aq.magnitude.dtype == np.complex128

    def test_real_property(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 2j, 3 + 4j], "meter", ureg)
        result = aq.real
        np.testing.assert_array_equal(result.magnitude, [1, 3])

    def test_imag_property(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 2j, 3 + 4j], "meter", ureg)
        result = aq.imag
        np.testing.assert_array_equal(result.magnitude, [2, 4])

    def test_conjugate(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 2j, 3 + 4j], "meter", ureg)
        result = ArrayQuantity(np.conjugate(aq.magnitude), "meter", ureg)
        np.testing.assert_array_equal(result.magnitude, [1 - 2j, 3 - 4j])

    def test_isreal(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 0j, 3 + 4j], "meter", ureg)
        result = np.isreal(aq)
        np.testing.assert_array_equal(result, [True, False])

    def test_iscomplex(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1 + 0j, 3 + 4j], "meter", ureg)
        result = np.iscomplex(aq)
        np.testing.assert_array_equal(result, [False, True])


class TestArrayQuantityToCompact:
    """Tests for to_compact on array quantities."""

    def test_to_compact_large(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([5000, 10000], "meter", ureg)
        result = aq.to_compact()
        assert isinstance(result, ArrayQuantity)
        assert len(result.magnitude) == 2
        # The magnitude should be rescaled to a compact unit
        assert result.magnitude[0] < 5000 or result.magnitude[1] < 10000

    def test_to_compact_small(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([0.001, 0.002], "meter", ureg)
        result = aq.to_compact()
        assert isinstance(result, ArrayQuantity)
        assert len(result.magnitude) == 2


class TestArrayQuantityCompatibility:
    """Tests for is_compatible_with on array quantities."""

    def test_compatible_same_dimension(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert aq.is_compatible_with("kilometer")

    def test_compatible_string(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert aq.is_compatible_with("centimeter")

    def test_incompatible(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert not aq.is_compatible_with("second")

    def test_compatible_with_array_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        other = ArrayQuantity([3, 4], "kilometer", ureg)
        assert aq.is_compatible_with(other)

    def test_incompatible_with_array_quantity(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        other = ArrayQuantity([3, 4], "second", ureg)
        assert not aq.is_compatible_with(other)


class TestArrayQuantityProperties:
    """Tests for various properties (shape, ndim, dtype, T, flat)."""

    def test_ndim(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.zeros((2, 3)), "meter", ureg)
        assert aq.ndim == 2

    def test_ndim_1d(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        assert aq.ndim == 1

    def test_T_property(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = aq.T
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [[1, 3], [2, 4]])

    def test_flat_property(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        flat_vals = list(aq.flat)
        assert flat_vals == [1, 2, 3, 4]

    def test_dimensionality(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert aq.dimensionality is not None

    def test_dimensionless_true(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "dimensionless", ureg)
        assert aq.dimensionless

    def test_dimensionless_false(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "meter", ureg)
        assert not aq.dimensionless


class TestArrayQuantityInPlaceConversion:
    """Tests for in-place conversion (ito)."""

    def test_ito(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1000, 2000], "meter", ureg)
        aq.ito("kilometer")
        np.testing.assert_allclose(aq.magnitude, [1, 2])
        assert "kilometer" in aq.units_str or "km" in aq.units_str

    def test_m_as(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2], "kilometer", ureg)
        result = aq.m_as("meter")
        np.testing.assert_allclose(result, [1000, 2000])


class TestArrayQuantityMiscUfuncs:
    """Tests for misc ufuncs: abs, negative, floor_divide, remainder."""

    def test_abs_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([-1, 2, -3], "meter", ureg)
        result = np.abs(a)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 3])

    def test_negative_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, -2, 3], "meter", ureg)
        result = np.negative(a)
        np.testing.assert_array_equal(result.magnitude, [-1, 2, -3])

    def test_floor_divide(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([7, 8, 9], "meter", ureg)
        b = ArrayQuantity([2, 3, 4], "second", ureg)
        result = np.floor_divide(a, b)
        np.testing.assert_array_equal(result.magnitude, [3, 2, 2])

    def test_remainder(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([7, 8, 9], "meter", ureg)
        b = ArrayQuantity([3, 3, 4], "meter", ureg)
        result = np.remainder(a, b)
        np.testing.assert_array_equal(result.magnitude, [1, 2, 1])

    def test_reciprocal(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1.0, 2.0, 4.0], "meter", ureg)
        result = np.reciprocal(a)
        np.testing.assert_allclose(result.magnitude, [1.0, 0.5, 0.25])

    def test_multiply_ufunc_two_quantities(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 3], "meter", ureg)
        b = ArrayQuantity([4, 5], "second", ureg)
        result = np.multiply(a, b)
        np.testing.assert_array_equal(result.magnitude, [8, 15])

    def test_divide_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([10, 20], "meter", ureg)
        b = ArrayQuantity([2, 5], "second", ureg)
        result = np.divide(a, b)
        np.testing.assert_array_equal(result.magnitude, [5, 4])

    def test_signbit_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([-1.0, 0.0, 1.0], "meter", ureg)
        result = np.signbit(a)
        np.testing.assert_array_equal(result, [True, False, False])

    def test_hypot_ufunc(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([3.0], "meter", ureg)
        b = ArrayQuantity([4.0], "meter", ureg)
        result = np.hypot(a, b)
        np.testing.assert_allclose(result.magnitude, [5.0])


class TestArrayQuantityReshapeOps:
    """Tests for np.reshape, np.sort, np.concatenate via __array_function__."""

    def test_np_reshape(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = np.reshape(aq, (2, 2))
        assert isinstance(result, ArrayQuantity)
        assert result.shape == (2, 2)
        np.testing.assert_array_equal(result.magnitude, [[1, 2], [3, 4]])

    def test_ones_like_via_function(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.ones_like(aq)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [1, 1, 1])

    def test_zeros_like_via_function(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.zeros_like(aq)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [0, 0, 0])

    def test_full_like_via_function(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = np.full_like(aq, 7)
        assert isinstance(result, ArrayQuantity)
        np.testing.assert_array_equal(result.magnitude, [7, 7, 7])


class TestArrayQuantityAddSubtractUnitChecking:
    """Tests for add/subtract with incompatible units raising errors."""

    def test_add_incompatible_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([1, 2], "second", ureg)
        with pytest.raises(DimensionalityError):
            np.add(a, b)

    def test_subtract_incompatible_raises(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([1, 2], "second", ureg)
        with pytest.raises(DimensionalityError):
            np.subtract(a, b)

    def test_add_compatible_converts(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([1, 2], "meter", ureg)
        b = ArrayQuantity([100, 200], "centimeter", ureg)
        result = np.add(a, b)
        np.testing.assert_allclose(result.magnitude, [2, 4])

    def test_subtract_compatible_converts(self, ureg: UnitRegistry) -> None:
        a = ArrayQuantity([2, 4], "meter", ureg)
        b = ArrayQuantity([100, 200], "centimeter", ureg)
        result = np.subtract(a, b)
        np.testing.assert_allclose(result.magnitude, [1, 2])


class TestArrayQuantityMedianPercentile:
    """Tests for np.median and np.percentile."""

    def test_median(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 3, 2, 5, 4], "meter", ureg)
        result = np.median(aq)
        assert float(result.magnitude) == pytest.approx(3.0)

    def test_percentile(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = np.percentile(aq, 50)
        assert float(result.magnitude) == pytest.approx(2.5)

    def test_percentile_25(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3, 4], "meter", ureg)
        result = np.percentile(aq, 25)
        assert float(result.magnitude) == pytest.approx(1.75)


class TestArrayQuantityDtypeConversion:
    """Tests for dtype and tolist."""

    def test_int_dtype(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([1, 2, 3], dtype=np.int32), "meter", ureg)
        assert aq.dtype == np.int32

    def test_float32_dtype(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([1, 2, 3], dtype=np.float32), "meter", ureg)
        assert aq.dtype == np.float32

    def test_tolist(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1, 2, 3], "meter", ureg)
        result = aq.tolist()
        assert result == [1, 2, 3]

    def test_array_protocol(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity([1.0, 2.0, 3.0], "meter", ureg)
        arr = np.asarray(aq)
        np.testing.assert_array_equal(arr, [1.0, 2.0, 3.0])


class TestArrayQuantityReductionAxes:
    """Tests for reduction operations with axis parameter."""

    def test_sum_axis0(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = np.sum(aq.magnitude, axis=0)
        np.testing.assert_array_equal(result, [4, 6])

    def test_sum_axis1(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = np.sum(aq.magnitude, axis=1)
        np.testing.assert_array_equal(result, [3, 7])

    def test_mean_axis(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = np.mean(aq.magnitude, axis=0)
        np.testing.assert_allclose(result, [2, 3])

    def test_std_axis(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = np.std(aq.magnitude, axis=0)
        np.testing.assert_allclose(result, [1, 1])

    def test_var_axis(self, ureg: UnitRegistry) -> None:
        aq = ArrayQuantity(np.array([[1, 2], [3, 4]]), "meter", ureg)
        result = np.var(aq.magnitude, axis=0)
        np.testing.assert_allclose(result, [1, 1])



"""Testing utilities for pintrs.

Provides assertion helpers for comparing unit-aware quantities,
similar to pint.testing.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pintrs._core import Quantity


def assert_equal(first: Quantity, second: Quantity) -> None:
    """Assert that two Quantities are exactly equal (magnitude and units).

    Args:
        first: First quantity.
        second: Second quantity.

    Raises:
        AssertionError: If quantities are not equal.
    """
    try:
        if str(first.units) != str(second.units):
            second = second.to(str(first.units))
    except Exception as exc:
        msg = f"{first} != {second}"
        raise AssertionError(msg) from exc
    if first.magnitude != second.magnitude:
        msg = f"{first} != {second}"
        raise AssertionError(msg)


def assert_allclose(
    first: Quantity,
    second: Quantity,
    rtol: float = 1e-7,
    atol: float = 0.0,
) -> None:
    """Assert that two Quantities are approximately equal.

    Converts to the same units before comparing magnitudes.

    Args:
        first: First quantity.
        second: Second quantity.
        rtol: Relative tolerance.
        atol: Absolute tolerance.

    Raises:
        AssertionError: If quantities are not close enough.
    """
    try:
        if str(first.units) != str(second.units):
            second = second.to(str(first.units))
    except Exception as exc:
        msg = f"{first} != {second} within rtol={rtol}, atol={atol}"
        raise AssertionError(msg) from exc
    a = first.magnitude
    b = second.magnitude
    if abs(a - b) > atol + rtol * abs(b):
        msg = (
            f"{first} != {second} within rtol={rtol}, atol={atol} "
            f"(diff={abs(a - b):.6e})"
        )
        raise AssertionError(msg)



"""Application registry tests - ported from pint's test_application_registry.py."""

from __future__ import annotations

import pickle

import pytest
from pintrs import (
    Measurement,
    UnitRegistry,
    get_application_registry,
    set_application_registry,
)


@pytest.fixture(autouse=True)
def _reset_app_registry() -> None:  # type: ignore[misc]
    """Reset application registry between tests."""
    import pintrs

    old = pintrs._application_registry
    yield  # type: ignore[misc]
    pintrs._application_registry = old


class TestDefaultApplicationRegistry:
    def test_get_application_registry(self) -> None:
        ureg = get_application_registry()
        assert ureg is not None
        u = ureg.parse_units("kg")
        assert str(u) in {"kg", "kilogram"}

    def test_get_returns_same_instance(self) -> None:
        ureg1 = get_application_registry()
        ureg2 = get_application_registry()
        assert ureg1 is ureg2

    def test_set_application_registry(self) -> None:
        new_ureg = UnitRegistry()
        set_application_registry(new_ureg)
        assert get_application_registry() is new_ureg

    def test_quantity_operations_with_app_registry(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(1, "meter")
        assert q.to("cm").magnitude == pytest.approx(100.0)

    @pytest.mark.parametrize("protocol", range(pickle.HIGHEST_PROTOCOL + 1))
    def test_pickle_quantity(self, protocol: int) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(123, "kg")
        restored = pickle.loads(pickle.dumps(q, protocol))
        assert restored.magnitude == pytest.approx(123.0)
        assert str(restored.units) in {"kg", "kilogram"}

    @pytest.mark.parametrize("protocol", range(pickle.HIGHEST_PROTOCOL + 1))
    def test_pickle_unit(self, protocol: int) -> None:
        ureg = get_application_registry()
        u = ureg.parse_units("kg")
        restored = pickle.loads(pickle.dumps(u, protocol))
        assert str(restored) in {"kg", "kilogram"}


class TestCustomApplicationRegistry:
    def test_custom_unit_in_app_registry(self) -> None:
        ureg = UnitRegistry()
        ureg.define("foo = [test_dim]")
        ureg.define("foo_half = foo / 2")
        set_application_registry(ureg)

        app = get_application_registry()
        q = app.Quantity(1, "foo")
        assert q.to("foo_half").magnitude == pytest.approx(2.0)

    def test_swap_registry(self) -> None:
        ureg1 = UnitRegistry()
        ureg2 = UnitRegistry()

        set_application_registry(ureg1)
        assert get_application_registry() is ureg1

        set_application_registry(ureg2)
        assert get_application_registry() is ureg2


class TestMeasurementWithRegistry:
    def test_measurement_from_registry(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(10, "meter")
        m = Measurement(q, 0.5)
        assert m.value.magnitude == pytest.approx(10.0)
        assert m.error.magnitude == pytest.approx(0.5)

    def test_measurement_str(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(10, "meter")
        m = Measurement(q, 0.5)
        s = str(m)
        assert "10" in s
        assert "0.5" in s


class TestRegistryIsolation:
    """Tests that registries are independent."""

    def test_multiple_registries_independent(self) -> None:
        ureg1 = UnitRegistry()
        ureg2 = UnitRegistry()
        q1 = ureg1.Quantity(1, "meter")
        q2 = ureg2.Quantity(1, "meter")
        assert q1.magnitude == q2.magnitude
        assert str(q1.units) == str(q2.units)

    def test_custom_unit_only_in_own_registry(self) -> None:
        ureg1 = UnitRegistry()
        ureg1.define("wobble = [dim_wobble]")
        q = ureg1.Quantity(1, "wobble")
        assert q.magnitude == 1.0

        ureg2 = UnitRegistry()
        with pytest.raises(Exception):
            ureg2.Quantity(1, "wobble")

    def test_set_restores_previous(self) -> None:
        original = get_application_registry()
        new_ureg = UnitRegistry()
        set_application_registry(new_ureg)
        assert get_application_registry() is new_ureg
        set_application_registry(original)
        assert get_application_registry() is original


class TestRegistryConversions:
    """Tests for quantity operations through the application registry."""

    def test_unit_conversion_kg_to_g(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(1, "kg")
        result = q.to("g")
        assert result.magnitude == pytest.approx(1000.0)

    def test_unit_conversion_m_to_km(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(1500, "meter")
        result = q.to("kilometer")
        assert result.magnitude == pytest.approx(1.5)

    def test_quantity_arithmetic(self) -> None:
        ureg = get_application_registry()
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(2, "meter")
        result = q1 + q2
        assert result.magnitude == pytest.approx(3.0)

    def test_quantity_multiplication(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(3, "meter")
        result = q * 2
        assert result.magnitude == pytest.approx(6.0)

    def test_incompatible_addition_raises(self) -> None:
        ureg = get_application_registry()
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(1, "second")
        with pytest.raises(Exception):
            _ = q1 + q2


class TestSwapApplicationRegistry:
    """Tests for swapping the application registry."""

    def test_swap_preserves_quantities(self) -> None:
        ureg1 = UnitRegistry()
        ureg2 = UnitRegistry()
        set_application_registry(ureg1)
        q1 = get_application_registry().Quantity(10, "meter")
        set_application_registry(ureg2)
        q2 = get_application_registry().Quantity(10, "meter")
        assert q1.magnitude == q2.magnitude

    def test_swap_custom_units(self) -> None:
        ureg1 = UnitRegistry()
        ureg1.define("flurp = [dim_flurp]")
        ureg1.define("flurp_half = flurp / 2")
        set_application_registry(ureg1)
        app = get_application_registry()
        q = app.Quantity(1, "flurp")
        assert q.to("flurp_half").magnitude == pytest.approx(2.0)

    def test_swap_back_to_default(self) -> None:
        default = get_application_registry()
        custom = UnitRegistry()
        set_application_registry(custom)
        assert get_application_registry() is custom
        set_application_registry(default)
        assert get_application_registry() is default


class TestPickleWithCustomRegistry:
    """Pickle tests with custom registries."""

    @pytest.mark.parametrize("protocol", range(pickle.HIGHEST_PROTOCOL + 1))
    def test_pickle_quantity_custom_registry(self, protocol: int) -> None:
        ureg = UnitRegistry()
        q = ureg.Quantity(42, "meter")
        restored = pickle.loads(pickle.dumps(q, protocol))
        assert restored.magnitude == pytest.approx(42.0)

    def test_pickle_preserves_units(self) -> None:
        ureg = get_application_registry()
        q = ureg.Quantity(3.14, "kilogram")
        restored = pickle.loads(pickle.dumps(q))
        assert str(restored.units) in {"kg", "kilogram"}
        assert restored.magnitude == pytest.approx(3.14)



"""Converter tests - ported from pint's test_converters.py.

Tests conversion behavior (scale, offset, logarithmic) through the
quantity API rather than through pint's internal converter classes.
"""

from __future__ import annotations

import math

import pytest
from pintrs import UnitRegistry


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestScaleConversion:
    """Test multiplicative/scale conversions (mirrors ScaleConverter tests)."""

    def test_simple_scale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "kilometer")
        result = q.to("meter")
        assert abs(result.magnitude - 1000) < 1e-10

    def test_roundtrip_scale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "centimeter")
        roundtrip = q.to("meter").to("centimeter")
        assert abs(roundtrip.magnitude - 100) < 1e-10

    def test_prefix_scale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "milligram")
        result = q.to("gram")
        assert abs(result.magnitude - 0.001) < 1e-10

    def test_micro_prefix(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1e6, "micrometer")
        result = q.to("meter")
        assert abs(result.magnitude - 1.0) < 1e-10

    def test_compound_scale(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(1, "kilometer / hour")
        result = q.to("meter / second")
        assert abs(result.magnitude - (1000 / 3600)) < 1e-10


class TestOffsetConversion:
    """Test offset conversions like temperature (mirrors OffsetConverter tests)."""

    def test_celsius_to_kelvin(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "degree_Celsius")
        result = q.to("kelvin")
        assert abs(result.magnitude - 273.15) < 1e-6

    def test_kelvin_to_celsius(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(273.15, "kelvin")
        result = q.to("degree_Celsius")
        assert abs(result.magnitude) < 1e-6

    def test_celsius_to_fahrenheit(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "degree_Celsius")
        result = q.to("degree_Fahrenheit")
        assert abs(result.magnitude - 212.0) < 1e-6

    def test_fahrenheit_to_celsius(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(32, "degree_Fahrenheit")
        result = q.to("degree_Celsius")
        assert abs(result.magnitude) < 1e-6

    def test_roundtrip_celsius_kelvin(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(100, "degree_Celsius")
        roundtrip = q.to("kelvin").to("degree_Celsius")
        assert abs(roundtrip.magnitude - 100) < 1e-6

    def test_roundtrip_fahrenheit_celsius(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(72, "degree_Fahrenheit")
        roundtrip = q.to("degree_Celsius").to("degree_Fahrenheit")
        assert abs(roundtrip.magnitude - 72) < 1e-6

    def test_absolute_zero_kelvin(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "kelvin")
        result = q.to("degree_Celsius")
        assert abs(result.magnitude - (-273.15)) < 1e-6

    def test_absolute_zero_fahrenheit(self, ureg: UnitRegistry) -> None:
        q = ureg.Quantity(0, "kelvin")
        result = q.to("degree_Fahrenheit")
        assert abs(result.magnitude - (-459.67)) < 1e-4


class TestLogarithmicConversion:
    """Test logarithmic conversions (mirrors LogarithmicConverter tests)."""

    def test_dBm_0_is_1mW(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(0, "dBm")
        linear = lq.to_linear()
        assert linear == pytest.approx(1.0)

    def test_dBm_30_is_1W(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(30, "dBm")
        linear = lq.to_linear()
        assert linear == pytest.approx(1000.0)

    def test_dB_0_is_unity(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(0, "dB")
        assert lq.to_linear() == pytest.approx(1.0)

    def test_dB_10_is_10(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(10, "dB")
        assert lq.to_linear() == pytest.approx(10.0)

    def test_dB_20_is_100(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(20, "dB")
        assert lq.to_linear() == pytest.approx(100.0)

    def test_dB_roundtrip(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        arb_value = 20.0
        lq = LogarithmicQuantity(arb_value, "dB")
        linear = lq.to_linear()
        back = LogarithmicQuantity.from_linear(linear, "dB")
        assert back.magnitude == pytest.approx(arb_value)

    def test_neper_to_linear(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(1, "Np")
        assert lq.to_linear() == pytest.approx(math.e**2)

    def test_bel_to_linear(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity(1, "B")
        assert lq.to_linear() == pytest.approx(10.0)

    def test_from_linear_dB(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity.from_linear(100, "dB")
        assert lq.magnitude == pytest.approx(20.0)

    def test_from_linear_dBm(self, ureg: UnitRegistry) -> None:
        from pintrs import LogarithmicQuantity

        lq = LogarithmicQuantity.from_linear(1, "dBm")
        assert lq.magnitude == pytest.approx(0.0)


class TestScaleConversionNumpy:
    """Test scale conversions with numpy arrays."""

    def test_array_scale_roundtrip(self, ureg: UnitRegistry) -> None:
        np = pytest.importorskip("numpy")
        from pintrs import ArrayQuantity

        a = ArrayQuantity(np.ones((1, 10)), "kilometer", ureg)
        result = a.to("meter").to("kilometer")
        np.testing.assert_allclose(result.magnitude, np.ones((1, 10)))



"""Unit formatting support for pintrs.

Provides pretty (Unicode), LaTeX, HTML, and compact formatting modes
compatible with pint's format specifiers.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pintrs._core import Quantity, Unit

_SUPERSCRIPT_MAP = str.maketrans("0123456789-", "⁰¹²³⁴⁵⁶⁷⁸⁹⁻")


def _format_default_magnitude(value: object) -> str:
    if isinstance(value, float) and value.is_integer():
        return str(int(value))
    return str(value)


def _parse_format_spec(spec: str) -> tuple[str, str, bool]:
    """Parse a pint-compatible format spec.

    Returns:
        Tuple of (magnitude_format, unit_mode, use_abbreviations).
        unit_mode is one of "D", "P", "L", "H", "C".
    """
    use_abbrev = "~" in spec
    clean = spec.replace("~", "")

    if clean.endswith("Lx"):
        unit_mode = "Lx"
        mag_fmt = clean[:-2]
    elif clean and clean[-1] in "PLHCD":
        unit_mode = clean[-1]
        mag_fmt = clean[:-1]
    else:
        unit_mode = "D"
        mag_fmt = clean

    return mag_fmt, unit_mode, use_abbrev


def _get_components(unit: Unit, use_abbrev: bool) -> list[tuple[str, float]]:
    """Get (display_name, exponent) pairs from a Unit.

    With use_abbrev=True, names come from _units_str (symbols).
    With use_abbrev=False, names come from _units_dict (canonical names).
    """
    raw = list(unit._units_dict())
    if not use_abbrev:
        return raw
    # Build a symbol mapping from the _units_str output
    symbol_str = unit._units_str()
    return _parse_unit_str_to_components(symbol_str, raw)


def _parse_unit_str_to_components(
    unit_str: str,
    raw: list[tuple[str, float]],
) -> list[tuple[str, float]]:
    """Parse _units_str output to extract symbol names with exponents.

    Falls back to raw components if parsing fails.
    """
    if not unit_str or unit_str == "dimensionless":
        return raw

    result: list[tuple[str, float]] = []
    # Split on / to get numerator and denominator parts
    parts = unit_str.split(" / ")
    numer_tokens = parts[0].split(" * ") if parts[0] != "1" else []
    denom_tokens = parts[1:] if len(parts) > 1 else []

    for token in numer_tokens:
        name, exp = _parse_unit_token(token)
        result.append((name, exp))
    for token in denom_tokens:
        name, exp = _parse_unit_token(token)
        result.append((name, -exp))

    return result if result else raw


def _parse_unit_token(token: str) -> tuple[str, float]:
    """Parse a single unit token like 's ** 2' into ('s', 2.0)."""
    token = token.strip()
    if " ** " in token:
        name, exp_str = token.rsplit(" ** ", 1)
        return name.strip(), float(exp_str)
    return token, 1.0


def _format_components(
    components: list[tuple[str, float]],
    *,
    exp_formatter: str = "default",
    separator: str = " * ",
    div_char: str = " / ",
) -> str:
    """Generic component formatter.

    Args:
        components: (name, exponent) pairs.
        exp_formatter: "default" for ** N, "pretty" for Unicode,
            "latex" for ^{N}, "html" for <sup>N</sup>.
        separator: Separator between units in numerator/denominator.
        div_char: Separator between numerator and denominator.
    """
    pos = [(n, e) for n, e in components if e > 0]
    neg = [(n, e) for n, e in components if e < 0]
    pos.sort(key=lambda x: x[0])
    neg.sort(key=lambda x: x[0])

    _exp_epsilon = 1e-9

    def _fmt(name: str, exp: float) -> str:
        a = abs(exp)
        if abs(a - 1.0) < _exp_epsilon:
            return name
        if exp_formatter == "pretty":
            s = str(int(a)) if a == int(a) else str(a)
            return name + s.translate(_SUPERSCRIPT_MAP)
        if exp_formatter == "latex":
            escaped = name.replace("_", r"\_")
            e_str = str(int(a)) if a == int(a) else str(a)
            return rf"\mathrm{{{escaped}}}^{{{e_str}}}"
        if exp_formatter == "html":
            e_str = str(int(a)) if a == int(a) else str(a)
            return f"{name}<sup>{e_str}</sup>"
        # default
        e_str = str(int(a)) if a == int(a) else str(a)
        return f"{name} ** {e_str}"

    if exp_formatter == "latex":
        pos_parts = [_fmt(n, e) for n, e in pos]
        neg_parts = [_fmt(n, abs(e)) for n, e in neg]
        numer = r" \cdot ".join(pos_parts) if pos_parts else "1"
        if not neg_parts:
            return rf"${numer}$"
        denom = r" \cdot ".join(neg_parts)
        return rf"$\frac{{{numer}}}{{{denom}}}$"

    pos_parts = [_fmt(n, e) for n, e in pos]
    neg_parts = [_fmt(n, abs(e)) for n, e in neg]

    if not neg_parts:
        return separator.join(pos_parts) if pos_parts else "dimensionless"
    if not pos_parts:
        return "1" + div_char + div_char.join(neg_parts)
    return div_char.join([separator.join(pos_parts), *neg_parts])


_SIUNITX_NAMES: dict[str, str] = {
    "m": "meter",
    "meter": "meter",
    "s": "second",
    "second": "second",
    "kg": "kilogram",
    "kilogram": "kilogram",
    "g": "gram",
    "gram": "gram",
    "A": "ampere",
    "ampere": "ampere",
    "K": "kelvin",
    "kelvin": "kelvin",
    "mol": "mole",
    "mole": "mole",
    "cd": "candela",
    "candela": "candela",
    "Hz": "hertz",
    "hertz": "hertz",
    "N": "newton",
    "newton": "newton",
    "Pa": "pascal",
    "pascal": "pascal",
    "J": "joule",
    "joule": "joule",
    "W": "watt",
    "watt": "watt",
    "C": "coulomb",
    "coulomb": "coulomb",
    "V": "volt",
    "volt": "volt",
    "F": "farad",
    "farad": "farad",
    "Ω": "ohm",
    "ohm": "ohm",
    "S": "siemens",
    "siemens": "siemens",
    "Wb": "weber",
    "weber": "weber",
    "T": "tesla",
    "tesla": "tesla",
    "H": "henry",
    "henry": "henry",
    "lm": "lumen",
    "lumen": "lumen",
    "lx": "lux",
    "lux": "lux",
    "Bq": "becquerel",
    "becquerel": "becquerel",
    "Gy": "gray",
    "gray": "gray",
    "Sv": "sievert",
    "sievert": "sievert",
    "L": "liter",
    "liter": "liter",
    "min": "minute",
    "minute": "minute",
    "h": "hour",
    "hour": "hour",
    "d": "day",
    "day": "day",
    "rad": "radian",
    "radian": "radian",
    "sr": "steradian",
    "steradian": "steradian",
}


_SI_PREFIXES: dict[str, str] = {
    "yocto": "yocto",
    "zepto": "zepto",
    "atto": "atto",
    "femto": "femto",
    "pico": "pico",
    "nano": "nano",
    "micro": "micro",
    "milli": "milli",
    "centi": "centi",
    "deci": "deci",
    "deca": "deca",
    "hecto": "hecto",
    "kilo": "kilo",
    "mega": "mega",
    "giga": "giga",
    "tera": "tera",
    "peta": "peta",
    "exa": "exa",
    "zetta": "zetta",
    "yotta": "yotta",
}


def _split_si_prefix(name: str) -> list[str]:
    """Split a unit name into prefix + base for siunitx."""
    for prefix in sorted(_SI_PREFIXES, key=len, reverse=True):
        if name.startswith(prefix) and name != prefix:
            base = name[len(prefix) :]
            if base in _SIUNITX_NAMES:
                return [rf"\{prefix}", rf"\{base}"]
    return [rf"\{name}"]


def _format_siunitx(components: list[tuple[str, float]]) -> str:
    """Format unit components in siunitx LaTeX style."""
    parts: list[str] = []
    for name, exp in sorted(components, key=lambda x: (-x[1], x[0])):
        si_name = _SIUNITX_NAMES.get(name, name)
        cmd_parts = _split_si_prefix(si_name)
        cmd = "".join(cmd_parts)
        if abs(exp - 1.0) < 1e-9:
            parts.append(cmd)
        elif abs(exp - (-1.0)) < 1e-9:
            parts.append(r"\per" + cmd)
        elif abs(exp - 2.0) < 1e-9:
            parts.append(cmd + r"\squared")
        elif abs(exp - 3.0) < 1e-9:
            parts.append(cmd + r"\cubed")
        elif abs(exp - (-2.0)) < 1e-9:
            parts.append(r"\per" + cmd + r"\squared")
        elif abs(exp - (-3.0)) < 1e-9:
            parts.append(r"\per" + cmd + r"\cubed")
        elif exp > 0:
            parts.append(cmd + rf"\tothe{{{exp}}}")
        else:
            parts.append(r"\per" + cmd + rf"\tothe{{{abs(exp)}}}")
    return "".join(parts)


def format_unit(unit: Unit, spec: str = "") -> str:
    """Format a Unit with pint-compatible format specifiers.

    Args:
        unit: The unit to format.
        spec: Format specifier. Supports ~, P, L, H, C, D suffixes.
    """
    _, unit_mode, use_abbrev = _parse_format_spec(spec)
    components = _get_components(unit, use_abbrev)

    if unit_mode == "P":
        return _format_components(
            components,
            exp_formatter="pretty",
            separator="·",
            div_char="/",
        )
    if unit_mode == "L":
        return _format_components(components, exp_formatter="latex")
    if unit_mode == "H":
        return _format_components(
            components,
            exp_formatter="html",
            separator=" ",
            div_char="/",
        )
    if unit_mode == "Lx":
        return _format_siunitx(components)
    # D or C
    return _format_components(components)


def format_quantity(quantity: Quantity, spec: str = "") -> str:
    """Format a Quantity with pint-compatible format specifiers.

    Args:
        quantity: The quantity to format.
        spec: Format specifier like ".2f~P", "~L", ".3g~H", etc.
    """
    mag_fmt, unit_mode, use_abbrev = _parse_format_spec(spec)

    q = quantity
    if unit_mode == "C":
        q = quantity.to_compact()

    mag_str = (
        format(q.magnitude, mag_fmt)
        if mag_fmt
        else _format_default_magnitude(
            q.magnitude,
        )
    )

    unit_spec = ("~" if use_abbrev else "") + unit_mode
    unit_str = format_unit(q.units, unit_spec)

    if unit_mode == "Lx":
        return rf"\SI[]{{{mag_str}}}{{{unit_str}}}"

    return f"{mag_str} {unit_str}"



"""Testing utilities tests - ported from pint's test_testing.py."""

from __future__ import annotations

import pytest

np = pytest.importorskip("numpy")

from pintrs import UnitRegistry
from pintrs.testing import assert_allclose, assert_equal


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestAssertEqual:
    def test_equal_quantities(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(1, "meter")
        assert_equal(q1, q2)

    def test_equal_after_conversion(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(100, "centimeter")
        assert_equal(q1, q2)

    def test_unequal_magnitudes_raises(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(2, "meter")
        with pytest.raises(AssertionError):
            assert_equal(q1, q2)

    def test_unequal_units_raises(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "meter")
        q2 = ureg.Quantity(1, "second")
        with pytest.raises(AssertionError):
            assert_equal(q1, q2)

    def test_equal_dimensionless(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1, "dimensionless")
        q2 = ureg.Quantity(1, "dimensionless")
        assert_equal(q1, q2)


class TestAssertAllclose:
    def test_close_quantities(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1.0, "meter")
        q2 = ureg.Quantity(1.0 + 1e-10, "meter")
        assert_allclose(q1, q2)

    def test_close_after_conversion(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1.0, "meter")
        q2 = ureg.Quantity(100.0, "centimeter")
        assert_allclose(q1, q2)

    def test_not_close_raises(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1.0, "meter")
        q2 = ureg.Quantity(2.0, "meter")
        with pytest.raises(AssertionError):
            assert_allclose(q1, q2)

    def test_custom_rtol(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1.0, "meter")
        q2 = ureg.Quantity(1.05, "meter")
        assert_allclose(q1, q2, rtol=0.1)

    def test_custom_atol(self, ureg: UnitRegistry) -> None:
        q1 = ureg.Quantity(1.0, "meter")
        q2 = ureg.Quantity(1.001, "meter")
        assert_allclose(q1, q2, atol=0.01)



"""Unit group support for pintrs.

Groups are named collections of units. For example, the "imperial" group
contains foot, inch, yard, mile, etc. Groups can include other groups
and support adding/removing units.
"""

# pyright: reportPrivateUsage=false, reportUnusedFunction=false

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pintrs._core import UnitRegistry as _UnitRegistry


class Group:
    """A named collection of units.

    Groups provide a way to organize units into logical sets (e.g., "imperial",
    "metric", "US_customary"). A special "root" group contains all units
    known to the registry.

    Args:
        name: Name of the group.
    """

    _REGISTRY: ClassVar[dict[str, Group]] = {}

    def __init__(self, name: str = "") -> None:
        self.name = name
        self._units: set[str] = set()
        self._used_groups: set[str] = set()
        if name:
            Group._REGISTRY[name] = self

    def add_units(self, *units: str) -> None:
        """Add units to this group.

        Args:
            units: Unit name strings to add.
        """
        self._units.update(units)

    def remove_units(self, *units: str) -> None:
        """Remove units from this group.

        Args:
            units: Unit name strings to remove.
        """
        self._units -= set(units)

    def add_used_groups(self, *groups: str) -> None:
        """Include other groups' units in this group.

        Args:
            groups: Group names to include.

        Raises:
            ValueError: If adding a group would create a cycle.
        """
        for gname in groups:
            if self._would_create_cycle(gname):
                msg = f"Adding group '{gname}' to '{self.name}' would create a cycle"
                raise ValueError(msg)
        self._used_groups.update(groups)

    def _would_create_cycle(self, candidate: str) -> bool:
        """Check if adding candidate group would create a circular dependency."""
        if candidate == self.name:
            return True
        visited: set[str] = set()
        return self._has_path_to(candidate, self.name, visited)

    @staticmethod
    def _has_path_to(start: str, target: str, visited: set[str]) -> bool:
        """Check if there's a path from start to target via used_groups."""
        if start in visited:
            return False
        visited.add(start)
        grp = Group._REGISTRY.get(start)
        if grp is None:
            return False
        for used in grp._used_groups:
            if used == target:
                return True
            if Group._has_path_to(used, target, visited):
                return True
        return False

    def remove_used_groups(self, *groups: str) -> None:
        """Stop including other groups' units.

        Args:
            groups: Group names to exclude.
        """
        self._used_groups -= set(groups)

    @property
    def members(self) -> frozenset[str]:
        """All units in this group, including from included groups."""
        result = set(self._units)
        for gname in self._used_groups:
            grp = Group._REGISTRY.get(gname)
            if grp is not None:
                result |= grp.members
        return frozenset(result)

    def __contains__(self, unit: str) -> bool:
        return unit in self.members

    def __iter__(self) -> Iterator[str]:
        return iter(self.members)

    def __len__(self) -> int:
        return len(self.members)

    def __repr__(self) -> str:
        return f"<Group('{self.name}', {len(self.members)} units)>"

    @staticmethod
    def get(name: str) -> Group | None:
        """Look up a group by name."""
        return Group._REGISTRY.get(name)

    @classmethod
    def from_lines(cls, lines: list[str]) -> Group:
        """Create a Group from definition lines.

        Args:
            lines: Lines from a group definition block.
        """
        name = lines[0].strip() if lines else ""
        grp = cls(name)
        for raw_line in lines[1:]:
            stripped = raw_line.strip()
            if not stripped or stripped.startswith("#"):
                continue
            for token in stripped.replace(",", " ").split():
                grp.add_units(token)
        return grp


_IMPERIAL_LENGTH = [
    "foot",
    "inch",
    "yard",
    "mile",
    "thou",
    "fathom",
    "furlong",
    "chain",
    "rod",
    "league",
    "hand",
    "link",
    "survey_foot",
    "survey_mile",
    "cables_length",
    "nautical_mile",
]
_IMPERIAL_MASS = [
    "pound",
    "ounce",
    "stone",
    "grain",
    "dram",
    "quarter",
    "hundredweight",
    "long_hundredweight",
    "long_ton",
    "UK_hundredweight",
    "UK_ton",
    "troy_ounce",
    "troy_pound",
    "pennyweight",
    "scruple",
    "apothecary_ounce",
    "apothecary_pound",
    "apothecary_dram",
    "carat",
    "slug",
    "slinch",
]
_IMPERIAL_VOLUME = [
    "imperial_gallon",
    "imperial_quart",
    "imperial_pint",
    "imperial_cup",
    "imperial_fluid_ounce",
    "imperial_gill",
    "imperial_barrel",
    "imperial_bushel",
    "imperial_peck",
]
_IMPERIAL_TEMP = ["degree_Fahrenheit", "degree_Rankine"]
_IMPERIAL_FORCE = [
    "force_pound",
    "force_ounce",
    "UK_force_ton",
    "US_force_ton",
    "poundal",
    "kip",
]
_IMPERIAL_PRESSURE = [
    "psi",
    "inch_Hg",
    "inch_Hg_32F",
    "inch_Hg_60F",
    "inch_H2O_39F",
    "inch_H2O_60F",
    "foot_Hg",
    "foot_Hg_0C",
    "foot_H2O",
    "foot_H2O_4C",
    "foot_H2O_60F",
]
_IMPERIAL_ENERGY = [
    "british_thermal_unit",
    "foot_pound",
]
_IMPERIAL_POWER = [
    "horsepower",
    "mechanical_horsepower",
    "electrical_horsepower",
    "boiler_horsepower",
]

_US_VOLUME = [
    "US_liquid_gallon",
    "US_liquid_quart",
    "US_liquid_pint",
    "US_liquid_cup",
    "US_fluid_ounce",
    "US_liquid_gill",
    "US_dry_gallon",
    "US_dry_quart",
    "US_dry_pint",
    "US_bushel",
    "US_peck",
    "US_dry_barrel",
    "US_liquid_barrel",
    "barrel",
]

_METRIC_LENGTH = ["meter"]
_METRIC_MASS = ["gram", "metric_ton"]
_METRIC_VOLUME = ["liter"]
_METRIC_TEMP = ["degree_Celsius", "kelvin"]
_METRIC_FORCE = ["newton", "dyne", "force_kilogram", "force_gram"]
_METRIC_PRESSURE = ["pascal", "bar", "atmosphere", "torr", "millimeter_Hg"]
_METRIC_ENERGY = ["joule", "calorie", "erg", "electron_volt"]
_METRIC_POWER = ["watt"]
_METRIC_TIME = ["second", "minute", "hour", "day"]


def _ensure_root_group(ureg: _UnitRegistry | None = None) -> Group:
    """Ensure the root group exists and return it."""
    existing = Group.get("root")
    if existing is not None:
        return existing
    root = Group("root")
    if ureg is not None:
        for unit_name in ureg:
            root.add_units(unit_name)
    return root


def _build_builtin_groups(ureg: _UnitRegistry | None = None) -> None:
    """Build built-in groups with well-known unit assignments.

    Args:
        ureg: Optional registry to validate unit names against.
    """
    root = _ensure_root_group(ureg)

    if "imperial" in Group._REGISTRY:
        return

    imperial = Group("imperial")
    all_imperial = (
        _IMPERIAL_LENGTH
        + _IMPERIAL_MASS
        + _IMPERIAL_VOLUME
        + _IMPERIAL_TEMP
        + _IMPERIAL_FORCE
        + _IMPERIAL_PRESSURE
        + _IMPERIAL_ENERGY
        + _IMPERIAL_POWER
    )
    valid_units = root.members if ureg is not None else None
    for u in all_imperial:
        if valid_units is None or u in valid_units:
            imperial.add_units(u)

    us = Group("US_customary")
    us.add_used_groups("imperial")
    for u in _US_VOLUME:
        if valid_units is None or u in valid_units:
            us.add_units(u)

    metric = Group("metric")
    all_metric = (
        _METRIC_LENGTH
        + _METRIC_MASS
        + _METRIC_VOLUME
        + _METRIC_TEMP
        + _METRIC_FORCE
        + _METRIC_PRESSURE
        + _METRIC_ENERGY
        + _METRIC_POWER
        + _METRIC_TIME
    )
    for u in all_metric:
        if valid_units is None or u in valid_units:
            metric.add_units(u)

    cgs = Group("cgs")
    cgs_units = [
        "centimeter",
        "gram",
        "second",
        "dyne",
        "erg",
        "barye",
        "poise",
        "stokes",
        "gauss",
        "oersted",
        "maxwell",
        "stilb",
        "phot",
        "gal",
    ]
    for u in cgs_units:
        if valid_units is None or u in valid_units:
            cgs.add_units(u)



"""Diskcache integration tests - ported from pint's test_diskcache.py.

Tests that pintrs Quantity/Unit objects can be cached via diskcache.
"""

from __future__ import annotations

import pytest
from pintrs import UnitRegistry

pytest.importorskip("diskcache", reason="diskcache is not available")
import diskcache


@pytest.fixture
def ureg() -> UnitRegistry:
    return UnitRegistry()


class TestDiskcacheQuantity:
    """Test caching pintrs Quantity objects with diskcache."""

    def test_cache_scalar_quantity(self, ureg: UnitRegistry, tmp_path: object) -> None:
        cache = diskcache.Cache(str(tmp_path))
        q = ureg.Quantity(42.0, "meter")
        cache["distance"] = q
        retrieved = cache["distance"]
        assert retrieved.magnitude == pytest.approx(42.0)
        assert str(retrieved.units) == str(q.units)
        cache.close()

    def test_cache_converted_quantity(
        self,
        ureg: UnitRegistry,
        tmp_path: object,
    ) -> None:
        cache = diskcache.Cache(str(tmp_path))
        q = ureg.Quantity(1000, "meter").to("kilometer")
        cache["dist_km"] = q
        retrieved = cache["dist_km"]
        assert retrieved.magnitude == pytest.approx(1.0)
        cache.close()

    def test_cache_multiple_quantities(
        self,
        ureg: UnitRegistry,
        tmp_path: object,
    ) -> None:
        cache = diskcache.Cache(str(tmp_path))
        cache["mass"] = ureg.Quantity(5.0, "kilogram")
        cache["time"] = ureg.Quantity(10.0, "second")
        cache["speed"] = ureg.Quantity(3.0, "meter / second")

        assert cache["mass"].magnitude == pytest.approx(5.0)
        assert cache["time"].magnitude == pytest.approx(10.0)
        assert cache["speed"].magnitude == pytest.approx(3.0)
        cache.close()

    def test_cache_overwrite(self, ureg: UnitRegistry, tmp_path: object) -> None:
        cache = diskcache.Cache(str(tmp_path))
        cache["val"] = ureg.Quantity(1.0, "meter")
        cache["val"] = ureg.Quantity(2.0, "meter")
        assert cache["val"].magnitude == pytest.approx(2.0)
        cache.close()

    def test_cache_no_cache_folder(self) -> None:
        ureg = UnitRegistry(cache_folder=None)
        assert ureg.cache_folder is None


class TestDiskcacheUnit:
    """Test caching pintrs Unit objects with diskcache."""

    def test_cache_unit(self, ureg: UnitRegistry, tmp_path: object) -> None:
        cache = diskcache.Cache(str(tmp_path))
        u = ureg.Unit("meter / second ** 2")
        cache["accel_unit"] = u
        retrieved = cache["accel_unit"]
        assert str(retrieved) == str(u)
        cache.close()



"""Type stubs for the pintrs package."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

from pintrs._core import (
    DefinitionSyntaxError as DefinitionSyntaxError,
)
from pintrs._core import (
    DimensionalityError as DimensionalityError,
)
from pintrs._core import (
    OffsetUnitCalculusError as OffsetUnitCalculusError,
)
from pintrs._core import (
    PintError as PintError,
)
from pintrs._core import (
    Quantity as Quantity,
)
from pintrs._core import (
    RedefinitionError as RedefinitionError,
)
from pintrs._core import (
    UndefinedUnitError as UndefinedUnitError,
)
from pintrs._core import (
    Unit as Unit,
)
from pintrs._core import (
    UnitRegistry as UnitRegistry,
)
from pintrs.context import Context as Context
from pintrs.group import Group as Group
from pintrs.logarithmic import LogarithmicQuantity as LogarithmicQuantity
from pintrs.numpy_support import ArrayQuantity as ArrayQuantity
from pintrs.system import System as System

__version__: str
__all__: list[str]

def set_application_registry(ureg: UnitRegistry) -> None: ...
def get_application_registry() -> UnitRegistry: ...
def wraps(
    ureg: UnitRegistry,
    ret: str | Unit | None,
    args: tuple[str | Unit | None, ...] | None = None,
    strict: bool = True,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
def check(
    ureg: UnitRegistry,
    *args: str | None,
) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
def make_quantity(
    ureg: UnitRegistry,
    value: Any,
    units: str | None = None,
) -> Any: ...

class Measurement:
    def __init__(
        self,
        value: Quantity | float,
        error: Quantity | float,
        units: str | None = None,
    ) -> None: ...
    @property
    def value(self) -> Quantity: ...
    @property
    def error(self) -> Quantity: ...
    @property
    def magnitude(self) -> float: ...
    @property
    def units(self) -> Unit: ...
    @property
    def rel(self) -> float: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __add__(self, other: Measurement) -> Measurement: ...
    def __sub__(self, other: Measurement) -> Measurement: ...
    def __mul__(self, other: Measurement | float) -> Measurement: ...
    def __truediv__(self, other: Measurement | float) -> Measurement: ...



"""Context-based conversions for pintrs.

Contexts allow converting between otherwise incompatible units by defining
custom transformation functions. For example, the spectroscopy context
allows converting between wavelength, frequency, and energy.
"""

from __future__ import annotations

import contextlib
from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    from collections.abc import Callable, Generator

    from pintrs._core import Quantity as ScalarQuantity
    from pintrs._core import UnitRegistry as _UnitRegistry

_PLANCK = 6.62607015e-34
_SPEED_OF_LIGHT = 299792458.0
_K_BOLTZMANN = 1.380649e-23
_AVOGADRO = 6.02214076e23

_DIM_LENGTH = "[length]"
_DIM_FREQUENCY = "1 / [time]"
_DIM_ENERGY = "[length] ** 2 * [mass] / [time] ** 2"


class Context:
    """A context that enables additional unit conversions.

    Contexts allow converting between otherwise incompatible dimensions
    by registering transformation functions between dimension pairs.

    Args:
        name: Name of the context.
        defaults: Default parameter values for transformations.
    """

    _REGISTRY: ClassVar[dict[str, Context]] = {}

    _DIM_TO_SI_UNIT: ClassVar[dict[str, str]] = {
        "[length]": "meter",
        "[time]": "second",
        "[mass]": "kilogram",
        "[temperature]": "kelvin",
        "[current]": "ampere",
        "[substance]": "mole",
        "[luminosity]": "candela",
        "1 / [time]": "hertz",
        "[length] ** 2 * [mass] / [time] ** 2": "joule",
    }

    def __init__(
        self,
        name: str = "",
        defaults: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        self.name = name
        self._transforms: dict[tuple[str, str], Callable[..., Any]] = {}
        self._defaults: dict[str, Any] = dict(defaults or {})
        self._defaults.update(kwargs)
        if name:
            Context._REGISTRY[name] = self

    @property
    def transforms(self) -> dict[tuple[str, str], Callable[..., Any]]:
        """Registered transformations."""
        return dict(self._transforms)

    @property
    def defaults(self) -> dict[str, Any]:
        """Default parameters passed to transformations."""
        return dict(self._defaults)

    def add_transformation(
        self,
        src: str,
        dst: str,
        func: Callable[..., Any],
    ) -> None:
        """Register a transformation function between two dimensions.

        Args:
            src: Source dimensionality string (e.g. "[length]").
            dst: Destination dimensionality string (e.g. "1 / [time]").
            func: Callable(ureg, value, **kwargs) -> transformed value.
        """
        self._transforms[(src, dst)] = func

    def transform(
        self,
        ureg: _UnitRegistry,
        quantity: ScalarQuantity,
        src_dim: str,
        dst_dim: str,
    ) -> Any:
        """Apply a registered transformation.

        Returns:
            Transformed magnitude in SI base units of dst, or None if no
            transformation is registered for this pair.
        """
        func = self._transforms.get((str(src_dim), str(dst_dim)))
        if func is None:
            return None
        result = func(ureg, quantity, **self._defaults)
        if hasattr(result, "check") and result.check(str(dst_dim)):
            return result

        si_unit = self._DIM_TO_SI_UNIT.get(str(src_dim))
        if si_unit is not None:
            mag = float(quantity.m_as(si_unit))
        else:
            base = quantity.to_base_units()
            mag = float(base.magnitude)
        return func(ureg, mag, **self._defaults)

    @staticmethod
    def from_lines(
        lines: list[str],
        to_base_func: Any = None,  # noqa: ARG004
    ) -> Context:
        """Create a Context from definition lines."""
        name = lines[0].strip() if lines else ""
        return Context(name)

    @staticmethod
    def get(name: str) -> Context | None:
        """Look up a context by name."""
        return Context._REGISTRY.get(name)


def _build_spectroscopy_context() -> Context:
    """Build the spectroscopy context (wavelength <-> frequency <-> energy)."""
    ctx = Context("spectroscopy")

    def _len_to_freq(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return _ureg.Quantity(_SPEED_OF_LIGHT, "meter / second") / value

    def _freq_to_len(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return _ureg.Quantity(_SPEED_OF_LIGHT, "meter / second") / value

    def _len_to_energy(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        hc = _ureg.Quantity(_PLANCK * _SPEED_OF_LIGHT, "joule * meter")
        return hc / value

    def _energy_to_len(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        hc = _ureg.Quantity(_PLANCK * _SPEED_OF_LIGHT, "joule * meter")
        return hc / value

    def _freq_to_energy(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return _ureg.Quantity(_PLANCK, "joule / hertz") * value

    def _energy_to_freq(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return value / _ureg.Quantity(_PLANCK, "joule / hertz")

    ctx.add_transformation(_DIM_LENGTH, _DIM_FREQUENCY, _len_to_freq)
    ctx.add_transformation(_DIM_FREQUENCY, _DIM_LENGTH, _freq_to_len)
    ctx.add_transformation(_DIM_LENGTH, _DIM_ENERGY, _len_to_energy)
    ctx.add_transformation(_DIM_ENERGY, _DIM_LENGTH, _energy_to_len)
    ctx.add_transformation(_DIM_FREQUENCY, _DIM_ENERGY, _freq_to_energy)
    ctx.add_transformation(_DIM_ENERGY, _DIM_FREQUENCY, _energy_to_freq)

    return ctx


def _build_boltzmann_context() -> Context:
    """Build the Boltzmann context (energy <-> temperature)."""
    ctx = Context("Boltzmann")

    def _energy_to_temp(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return value / _ureg.Quantity(_K_BOLTZMANN, "joule / kelvin")

    def _temp_to_energy(_ureg: _UnitRegistry, value: ScalarQuantity, **_kw: Any) -> Any:
        return value * _ureg.Quantity(_K_BOLTZMANN, "joule / kelvin")

    ctx.add_transformation(_DIM_ENERGY, "[temperature]", _energy_to_temp)
    ctx.add_transformation("[temperature]", _DIM_ENERGY, _temp_to_energy)

    return ctx


def _build_chemistry_context() -> Context:
    """Build the chemistry context (amount <-> count via Avogadro)."""
    ctx = Context("chemistry")

    def _amount_to_count(
        _ureg: _UnitRegistry,
        value: ScalarQuantity,
        **_kw: Any,
    ) -> Any:
        return value * _AVOGADRO

    def _count_to_amount(
        _ureg: _UnitRegistry,
        value: ScalarQuantity,
        **_kw: Any,
    ) -> Any:
        return value / _AVOGADRO

    ctx.add_transformation("[substance]", "[dimensionless]", _amount_to_count)
    ctx.add_transformation("[dimensionless]", "[substance]", _count_to_amount)

    return ctx


# Build built-in contexts at import time
_build_spectroscopy_context()
_build_boltzmann_context()
_build_chemistry_context()


class ActiveContexts:
    """Manages which contexts are currently active on a registry."""

    def __init__(self) -> None:
        self._active: list[Context] = []

    @property
    def active(self) -> list[Context]:
        """Currently active contexts."""
        return list(self._active)

    def enable(self, *names: str | Context) -> None:
        """Enable one or more contexts by name or instance."""
        for name in names:
            if isinstance(name, Context):
                ctx: Context = name
            else:
                found = Context.get(name)
                if found is None:
                    msg = f"Unknown context: {name}"
                    raise KeyError(msg)
                ctx = found
            if ctx not in self._active:
                self._active.append(ctx)

    def disable(self, *names: str | Context) -> None:
        """Disable one or more contexts."""
        for name in names:
            if isinstance(name, Context):
                ctx: Context = name
            else:
                found = Context.get(name)
                if found is None:
                    continue
                ctx = found
            if ctx in self._active:
                self._active.remove(ctx)

    def clear(self) -> None:
        """Disable all contexts."""
        self._active.clear()

    @contextlib.contextmanager
    def __call__(
        self,
        *names: str | Context,
    ) -> Generator[None, None, None]:
        """Context manager that temporarily enables contexts."""
        prev = list(self._active)
        self.enable(*names)
        try:
            yield
        finally:
            self._active = prev

    def find_transform(
        self,
        ureg: _UnitRegistry,
        quantity: ScalarQuantity,
        src_dim: str,
        dst_dim: str,
    ) -> Any:
        """Search active contexts for a valid transformation."""
        for ctx in self._active:
            result = ctx.transform(ureg, quantity, src_dim, dst_dim)
            if result is not None:
                return result
        return None



"""Type stubs for the native Rust extension module."""

from __future__ import annotations

import datetime
from collections.abc import Callable, Iterator
from typing import Any

from pintrs.context import Context
from pintrs.group import Group
from pintrs.system import System

class PintError(ValueError): ...
class DimensionalityError(PintError): ...
class UndefinedUnitError(PintError): ...
class OffsetUnitCalculusError(PintError): ...
class DefinitionSyntaxError(PintError): ...
class RedefinitionError(PintError): ...

class Unit:
    def __init__(self, units: str) -> None: ...
    @property
    def dimensionality(self) -> str: ...
    @property
    def dimensionless(self) -> bool: ...
    def is_compatible_with(self, other: Unit | str) -> bool: ...
    def compatible_units(self) -> list[str]: ...
    def from_(self, quantity: Quantity) -> Quantity: ...
    def m_from(self, quantity: Quantity) -> float: ...
    @property
    def systems(self) -> list[str]: ...
    def compare(self, other: Unit, op: str) -> bool: ...
    def _units_str(self) -> str: ...
    def _units_dict(self) -> list[tuple[str, float]]: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __mul__(self, other: Any) -> Unit | Quantity: ...
    def __rmul__(self, other: Any) -> Unit | Quantity: ...
    def __truediv__(self, other: Any) -> Unit | Quantity: ...
    def __rtruediv__(self, other: Any) -> Quantity: ...
    def __pow__(self, exp: float) -> Unit: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Unit: ...
    def __deepcopy__(self, memo: object) -> Unit: ...
    def __reduce__(self) -> tuple[type, tuple[str]]: ...
    def format_babel(self, locale: str = "en") -> str: ...

class Quantity:
    def __init__(
        self,
        value: float | str | Quantity,
        units: str | None = None,
    ) -> None: ...
    @property
    def magnitude(self) -> float: ...
    @property
    def m(self) -> float: ...
    @property
    def units(self) -> Unit: ...
    @property
    def u(self) -> Unit: ...
    @property
    def dimensionality(self) -> str: ...
    @property
    def dimensionless(self) -> bool: ...
    @property
    def unitless(self) -> bool: ...
    def m_as(self, units: str) -> float: ...
    def to(self, units: str) -> Quantity: ...
    def ito(self, units: str) -> None: ...
    def to_base_units(self) -> Quantity: ...
    def to_root_units(self) -> Quantity: ...
    def ito_base_units(self) -> None: ...
    def ito_root_units(self) -> None: ...
    def to_compact(self) -> Quantity: ...
    def to_reduced_units(self) -> Quantity: ...
    def ito_reduced_units(self) -> None: ...
    def to_preferred(
        self,
        preferred_units: list[str] | None = None,
    ) -> Quantity: ...
    def ito_preferred(
        self,
        preferred_units: list[str] | None = None,
    ) -> None: ...
    def to_unprefixed(self) -> Quantity: ...
    def ito_unprefixed(self) -> None: ...
    def to_tuple(self) -> tuple[float, list[tuple[str, float]]]: ...
    @staticmethod
    def from_tuple(
        tup: tuple[float, list[tuple[str, float]]],
    ) -> Quantity: ...
    def unit_items(self) -> list[tuple[str, float]]: ...
    def compare(self, other: Quantity, op: str) -> bool: ...
    def to_timedelta(self) -> datetime.timedelta: ...
    def plus_minus(self, error: float) -> Any: ...
    def check(self, dimension: str) -> bool: ...
    def is_compatible_with(
        self,
        other: Quantity | Unit | str,
    ) -> bool: ...
    def compatible_units(self) -> list[str]: ...
    def _units_str(self) -> str: ...
    def _units_dict(self) -> list[tuple[str, float]]: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __format__(self, spec: str) -> str: ...
    def __float__(self) -> float: ...
    def __int__(self) -> int: ...
    def __bool__(self) -> bool: ...
    def __neg__(self) -> Quantity: ...
    def __abs__(self) -> Quantity: ...
    def __round__(self, ndigits: int | None = None) -> Quantity: ...
    def __add__(self, other: Quantity | float) -> Quantity: ...
    def __radd__(self, other: Quantity | float) -> Quantity: ...
    def __sub__(self, other: Quantity | float) -> Quantity: ...
    def __rsub__(self, other: Quantity | float) -> Quantity: ...
    def __mul__(self, other: Quantity | Unit | float) -> Quantity: ...
    def __rmul__(self, other: Quantity | Unit | float) -> Quantity: ...
    def __truediv__(
        self,
        other: Quantity | Unit | float,
    ) -> Quantity: ...
    def __rtruediv__(self, other: float) -> Quantity: ...
    def __pow__(self, exp: float) -> Quantity: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __lt__(self, other: Quantity) -> bool: ...
    def __le__(self, other: Quantity) -> bool: ...
    def __gt__(self, other: Quantity) -> bool: ...
    def __ge__(self, other: Quantity) -> bool: ...
    def __hash__(self) -> int: ...
    def __copy__(self) -> Quantity: ...
    def __deepcopy__(self, memo: object) -> Quantity: ...
    def __reduce__(self) -> tuple[type, tuple[float, str]]: ...
    # Monkey-patched numpy-like properties
    @property
    def ndim(self) -> int: ...
    @property
    def shape(self) -> tuple[int, ...]: ...
    @property
    def dtype(self) -> Any: ...
    @property
    def real(self) -> Quantity: ...
    @property
    def imag(self) -> Quantity: ...
    @property
    def T(self) -> Quantity: ...  # noqa: N802
    @property
    def flat(self) -> Iterator[Quantity]: ...
    force_ndarray: bool
    force_ndarray_like: bool
    def clip(self, min: Any = None, max: Any = None) -> Quantity: ...  # noqa: A002
    def dot(self, other: Any) -> Quantity: ...
    def prod(self, axis: Any = None) -> Quantity: ...
    def fill(self, value: Any) -> None: ...
    def put(self, indices: Any, values: Any) -> None: ...
    def searchsorted(self, v: Any, side: str = "left") -> Any: ...
    def tolist(self) -> Any: ...
    @classmethod
    def from_list(cls, lst: list[Any], units: str | None = None) -> Quantity: ...
    @classmethod
    def from_sequence(cls, seq: Any, units: str | None = None) -> Quantity: ...
    def compute(self, **kwargs: Any) -> Quantity: ...
    def persist(self, **kwargs: Any) -> Quantity: ...
    def visualize(self, **kwargs: Any) -> Any: ...
    def format_babel(self, locale: str = "en") -> str: ...
    UnitsContainer: Any

# Use _Q and _U aliases inside UnitRegistry to avoid mypy confusing
# the Quantity/Unit method names with the Quantity/Unit class types.
_Q = Quantity
_U = Unit

class UnitRegistry:
    def __init__(self) -> None: ...
    def Quantity(  # noqa: N802
        self,
        value: float | str | _Q | Any,
        units: str | None = None,
    ) -> _Q: ...
    def Unit(self, units: str) -> _U: ...  # noqa: N802
    def parse_expression(self, expr: str) -> _Q: ...
    def parse_units(self, units: str) -> _U: ...
    def convert(self, value: float, src: str, dst: str) -> float: ...
    def define(self, definition: str) -> None: ...
    def get_compatible_units(self, unit: str) -> list[str]: ...
    def get_base_units(self, unit: str) -> tuple[float, _U]: ...
    def get_root_units(self, unit: str) -> tuple[float, _U]: ...
    def get_dimensionality(self, unit: str) -> str: ...
    def get_name(self, unit: str) -> str: ...
    def get_symbol(self, unit: str) -> str: ...
    def is_compatible_with(self, unit1: str, unit2: str) -> bool: ...
    def parse_unit_name(
        self,
        unit: str,
    ) -> list[tuple[str, str, str]]: ...
    def load_definitions(self, definition: str) -> None: ...
    @property
    def default_format(self) -> str: ...
    @property
    def auto_reduce_dimensions(self) -> bool: ...
    def _scalar_quantity(self, value: Any, units: str | None = None) -> _Q: ...
    def _f64_quantity(self, value: float, units: str | None = None) -> _Q: ...
    def _get_conversion_factor(self, src: str, dst: str) -> float: ...
    def _get_root_units(self, unit: str) -> tuple[float, str]: ...
    def _get_dimensionality(self, unit: str) -> str: ...
    def _format_units(self, unit: str) -> str: ...
    def _get_prefixes(self) -> list[tuple[str, float]]: ...
    def __contains__(self, item: str) -> bool: ...
    def __iter__(self) -> Iterator[str]: ...
    def __getattr__(self, name: str) -> _Q: ...
    # Monkey-patched methods
    def wraps(
        self,
        ret: str | _U | None,
        args: tuple[str | _U | None, ...] | None = None,
        strict: bool = True,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
    def check(
        self,
        *args: str | None,
    ) -> Callable[[Callable[..., Any]], Callable[..., Any]]: ...
    def Measurement(  # noqa: N802
        self,
        value: float,
        error: float,
        units: str | None = None,
    ) -> Any: ...
    def setup_matplotlib(self, enable: bool = True) -> None: ...
    def context(self, *contexts: str | Context, **kwargs: Any) -> Any: ...
    def add_context(self, context: Context) -> None: ...
    def remove_context(self, name: str) -> None: ...
    def enable_contexts(self, *names: str, **kwargs: Any) -> None: ...
    def disable_contexts(self, *names: str) -> None: ...
    def with_context(self, *names: str, **kwargs: Any) -> Any: ...
    def get_group(self, name: str) -> Group: ...
    def get_system(self, name: str) -> System: ...
    def parse_units_as_container(self, units: str) -> dict[str, float]: ...
    def parse_pattern(self, pattern: str, input_string: str) -> Any: ...
    def pi_theorem(self, quantities: dict[str, str]) -> list[dict[str, float]]: ...
    def set_fmt_locale(self, locale: str | None) -> None: ...
    # Config flags
    autoconvert_offset_to_baseunit: bool
    autoconvert_to_preferred: bool
    cache_folder: str | None
    case_sensitive: bool
    default_as_delta: bool
    default_system: str
    fmt_locale: str | None
    force_ndarray: bool
    force_ndarray_like: bool
    non_int_type: type
    separate_format_defaults: bool | None
    preprocessors: list[Any]
    mpl_formatter: str
    # Class references
    Context: type[Context]
    Group: type[Group]
    System: type[System]
    UnitsContainer: Any
    @property
    def sys(self) -> Any: ...
    @property
    def formatter(self) -> Any: ...
